import copy
import re
from typing import Any, Dict, List, Optional, Union

from pytest import mark, param, raises

from omegaconf import (
    MISSING,
    Container,
    DictConfig,
    IntegerNode,
    KeyValidationError,
    ListConfig,
    Node,
    OmegaConf,
    ValidationError,
)
from omegaconf._utils import (
    ValueKind,
    _ensure_container,
    _resolve_optional,
    get_value_kind,
    is_dict_annotation,
    is_list_annotation,
    is_structured_config,
)
from tests import ConcretePlugin, Plugin


def check_node_metadata(
    node: Container,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
) -> None:
    value_optional, value_ref_type = _resolve_optional(type_hint)
    assert node._metadata.optional == value_optional
    assert node._metadata.ref_type == value_ref_type
    assert node._metadata.key_type == key_type
    assert node._metadata.element_type == elt_type
    assert node._metadata.object_type == obj_type

    if is_dict_annotation(value_ref_type) or is_structured_config(value_ref_type):
        assert isinstance(node, DictConfig)
    elif is_list_annotation(value_ref_type):
        assert isinstance(node, ListConfig)


def check_subnode(
    cfg: Container,
    key: Any,
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
) -> None:
    "Validate that `cfg[key] == value` and that subnode `cfg._get_node(key)._metadata` is correct."
    node = cfg._get_node(key)
    assert isinstance(node, (ListConfig, DictConfig))
    vk = get_value_kind(node)

    if vk in (ValueKind.MANDATORY_MISSING, ValueKind.INTERPOLATION):
        if isinstance(value, Node):
            value = value._value()
        assert node._value() == value
    else:
        assert cfg[key] == value

    check_node_metadata(node, type_hint, key_type, elt_type, obj_type)


@mark.parametrize(
    "cfg, type_hint, key_type, elt_type, obj_type",
    [
        param(
            ListConfig([[[456]]], element_type=List[List[int]]),
            List[List[int]],
            int,
            List[int],
            list,
            id="list-list-list",
        ),
        param(
            ListConfig([{"foo": {"bar": 456}}], element_type=Dict[str, Dict[str, int]]),
            Dict[str, Dict[str, int]],
            str,
            Dict[str, int],
            dict,
            id="list-dict-dict",
        ),
        param(
            ListConfig([[123], None], element_type=Optional[List[int]]),
            Optional[List[int]],
            int,
            int,
            list,
            id="list-optional-list",
        ),
        param(
            ListConfig([[123], [None]], element_type=List[Optional[int]]),
            List[Optional[int]],
            int,
            Optional[int],
            list,
            id="list-list-optional",
        ),
        param(
            ListConfig([{"bar": 456}, None], element_type=Optional[Dict[str, int]]),
            Optional[Dict[str, int]],
            str,
            int,
            dict,
            id="list-optional-dict",
        ),
        param(
            ListConfig(
                [{"foo": 456}, {"bar": None}], element_type=Dict[str, Optional[int]]
            ),
            Dict[str, Optional[int]],
            str,
            Optional[int],
            dict,
            id="list-dict-optional",
        ),
        param(
            DictConfig({"foo": [[456]]}, element_type=List[List[int]]),
            List[List[int]],
            int,
            List[int],
            list,
            id="dict-list-list",
        ),
        param(
            DictConfig(
                {"foo": {"bar": {"baz": 456}}}, element_type=Dict[str, Dict[str, int]]
            ),
            Dict[str, Dict[str, int]],
            str,
            Dict[str, int],
            dict,
            id="dict-dict-dict",
        ),
        param(
            DictConfig({"foo": [123], "bar": None}, element_type=Optional[List[int]]),
            Optional[List[int]],
            int,
            int,
            list,
            id="dict-optional-list",
        ),
        param(
            DictConfig({"foo": [123], "bar": [None]}, element_type=List[Optional[int]]),
            List[Optional[int]],
            int,
            Optional[int],
            list,
            id="dict-list-optional",
        ),
        param(
            DictConfig(
                {"foo": {"bar": 456}, "baz": None},
                element_type=Optional[Dict[str, int]],
            ),
            Optional[Dict[str, int]],
            str,
            int,
            dict,
            id="dict-optional-dict",
        ),
        param(
            DictConfig(
                {"foo": {"bar": 456}, "baz": {"qux": None}},
                element_type=Dict[str, Optional[int]],
            ),
            Dict[str, Optional[int]],
            str,
            Optional[int],
            dict,
            id="dict-dict-optional",
        ),
        param(
            DictConfig(
                {"foo": {"bar": ConcretePlugin()}}, element_type=Dict[str, Plugin]
            ),
            Dict[str, Plugin],
            str,
            Plugin,
            dict,
            id="dict-of-plugin",
        ),
        param(
            DictConfig({"foo": [ConcretePlugin()]}, element_type=List[Plugin]),
            List[Plugin],
            int,
            Plugin,
            list,
            id="list-of-plugin",
        ),
    ],
)
def test_container_nested_element(
    cfg: Union[DictConfig, ListConfig],
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
) -> None:
    """Ensure metadata and contents of container-typed subnode are correct"""
    cfg = copy.deepcopy(cfg)
    keys: Any = range(len(cfg)) if isinstance(cfg, ListConfig) else cfg.keys()
    for key in keys:
        value = cfg[key]
        check_subnode(
            cfg,
            key,
            value,
            type_hint,
            key_type,
            elt_type,
            obj_type if value is not None else None,
        )


@mark.parametrize(
    "cfg, value, type_hint, key_type, elt_type, obj_type",
    [
        param(
            ListConfig([[[456]]], element_type=List[List[int]]),
            [[123]],
            List[List[int]],
            int,
            List[int],
            list,
            id="assign-to-list-element",
        ),
        param(
            ListConfig([{"foo": {"bar": 456}}], element_type=Dict[str, Dict[str, int]]),
            {"baz": {"qux": 123}},
            Dict[str, Dict[str, int]],
            str,
            Dict[str, int],
            dict,
            id="assign-to-dict-element",
        ),
        param(
            ListConfig([[123], None], element_type=Optional[List[int]]),
            [456],
            Optional[List[int]],
            int,
            int,
            list,
            id="assign-list-to-optional-list",
        ),
        param(
            ListConfig([{"foo": 456}, None], element_type=Optional[Dict[str, int]]),
            {"bar": 123},
            Optional[Dict[str, int]],
            str,
            int,
            dict,
            id="assign-dict-to-optional-dict",
        ),
        param(
            ListConfig([[123], [None]], element_type=List[Optional[int]]),
            [456],
            List[Optional[int]],
            int,
            Optional[int],
            list,
            id="assign-list-to-list-optional",
        ),
        param(
            ListConfig([[123], [None]], element_type=List[Optional[int]]),
            [None],
            List[Optional[int]],
            int,
            Optional[int],
            list,
            id="assign-list-none-to-list-optional",
        ),
        param(
            ListConfig(
                [{"foo": 456}, {"bar": None}], element_type=Dict[str, Optional[int]]
            ),
            {"baz": 123},
            Dict[str, Optional[int]],
            str,
            Optional[int],
            dict,
            id="assign-dict-to-dict-optional",
        ),
        param(
            ListConfig(
                [{"foo": 456}, {"bar": None}], element_type=Dict[str, Optional[int]]
            ),
            {"baz": None},
            Dict[str, Optional[int]],
            str,
            Optional[int],
            dict,
            id="assign-dict-none-to-dict-optional",
        ),
        param(
            ListConfig([{"foo": ConcretePlugin()}], element_type=Dict[str, Plugin]),
            {"bar": ConcretePlugin()},
            Dict[str, Plugin],
            str,
            Plugin,
            dict,
            id="assign-dict-plugin",
        ),
        param(
            ListConfig([[ConcretePlugin()]], element_type=List[Plugin]),
            [ConcretePlugin()],
            List[Plugin],
            int,
            Plugin,
            list,
            id="assign-list-plugin",
        ),
    ],
)
@mark.parametrize(
    "ensure_container",
    [
        param(True, id="container"),
        param(False, id="no_container"),
    ],
)
def test_list_assign_to_container_typed_element(
    cfg: ListConfig,
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
    ensure_container: bool,
) -> None:
    cfg = copy.deepcopy(cfg)
    if ensure_container:
        value = _ensure_container(value)

    n = len(cfg)
    for idx in range(n):
        cfg[idx] = value
        check_subnode(cfg, idx, value, type_hint, key_type, elt_type, obj_type)

    cfg.append(value)
    check_subnode(cfg, n, value, type_hint, key_type, elt_type, obj_type)


@mark.parametrize(
    "cfg, type_hint, key_type, elt_type",
    [
        param(
            ListConfig([[123], None], element_type=Optional[List[int]]),
            Optional[List[int]],
            int,
            int,
            id="assign-to-optional-list",
        ),
        param(
            ListConfig([{"bar": 456}, None], element_type=Optional[Dict[str, int]]),
            Optional[Dict[str, int]],
            str,
            int,
            id="assign-to-optional-dict",
        ),
        param(
            ListConfig([[ConcretePlugin()], None], element_type=Optional[List[Plugin]]),
            Optional[List[Plugin]],
            int,
            Plugin,
            id="assign-to-optional-plugin-list",
        ),
        param(
            ListConfig(
                [{"bar": ConcretePlugin()}, None],
                element_type=Optional[Dict[str, Plugin]],
            ),
            Optional[Dict[str, Plugin]],
            str,
            Plugin,
            id="assign-to-optional-plugin-dict",
        ),
    ],
)
@mark.parametrize(
    "value",
    [
        param(None, id="none"),
        param(MISSING, id="missing"),
        param("${interp}", id="interp"),
    ],
)
def test_list_assign_to_container_typed_element_special(
    cfg: ListConfig,
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
) -> None:
    cfg = copy.deepcopy(cfg)

    n = len(cfg)
    for idx in range(n):
        cfg[idx] = value
        check_subnode(cfg, idx, value, type_hint, key_type, elt_type, None)

    cfg.append(value)
    check_subnode(cfg, n, value, type_hint, key_type, elt_type, None)


@mark.parametrize(
    "ensure_container",
    [
        param(True, id="container"),
        param(False, id="no_container"),
    ],
)
@mark.parametrize(
    "cfg, value, type_hint, key_type, elt_type, obj_type",
    [
        param(
            DictConfig({"foo": [[456]]}, element_type=List[List[int]]),
            [[123]],
            List[List[int]],
            int,
            List[int],
            list,
            id="assign-to-list-element",
        ),
        param(
            DictConfig(
                {"foo": {"bar": {"baz": 456}}}, element_type=Dict[str, Dict[str, int]]
            ),
            {"qux": {"frob": 123}},
            Dict[str, Dict[str, int]],
            str,
            Dict[str, int],
            dict,
            id="assign-to-dict-element",
        ),
        param(
            DictConfig({"foo": [123], "bar": None}, element_type=Optional[List[int]]),
            [456],
            Optional[List[int]],
            int,
            int,
            list,
            id="assign-list-to-optional-list",
        ),
        param(
            DictConfig(
                {"foo": {"bar": 456}, "baz": None},
                element_type=Optional[Dict[str, int]],
            ),
            {"qux": 123},
            Optional[Dict[str, int]],
            str,
            int,
            dict,
            id="assign-dict-to-optional-dict",
        ),
        param(
            DictConfig({"foo": [123], "bar": [None]}, element_type=List[Optional[int]]),
            [456],
            List[Optional[int]],
            int,
            Optional[int],
            list,
            id="assign-list-to-list-optional",
        ),
        param(
            DictConfig({"foo": [123], "bar": [None]}, element_type=List[Optional[int]]),
            [None],
            List[Optional[int]],
            int,
            Optional[int],
            list,
            id="assign-list-none-to-list-optional",
        ),
        param(
            DictConfig(
                {"foo": {"bar": 456}, "baz": {"qux": None}},
                element_type=Dict[str, Optional[int]],
            ),
            {"frob": 123},
            Dict[str, Optional[int]],
            str,
            Optional[int],
            dict,
            id="assign-dict-to-dict-optional",
        ),
        param(
            DictConfig(
                {"foo": {"bar": 456}, "baz": {"qux": None}},
                element_type=Dict[str, Optional[int]],
            ),
            {"frob": None},
            Dict[str, Optional[int]],
            str,
            Optional[int],
            dict,
            id="assign-dict-none-to-dict-optional",
        ),
        param(
            DictConfig({"foo": [ConcretePlugin()]}, element_type=List[Plugin]),
            [ConcretePlugin()],
            List[Plugin],
            int,
            Plugin,
            list,
            id="assign-to-list-of-plugins",
        ),
        param(
            DictConfig(
                {"foo": {"bar": ConcretePlugin()}}, element_type=Dict[str, Plugin]
            ),
            {"baz": ConcretePlugin()},
            Dict[str, Plugin],
            str,
            Plugin,
            dict,
            id="assign-to-dict-of-plugins",
        ),
        param(
            DictConfig({"key": []}, element_type=List[int]),
            DictConfig("${interp}"),
            List[int],
            int,
            int,
            None,
            id="coerce-dictconfig-interp-to-listconfig",
        ),
        param(
            DictConfig({"key": {}}, element_type=Dict[str, int]),
            ListConfig("${interp}"),
            Dict[str, int],
            str,
            int,
            None,
            id="coerce-listconfig-interp-to-dictconfig",
        ),
        param(
            DictConfig({"key": []}, element_type=List[int]),
            DictConfig("${interp}", ref_type=Dict[str, int]),
            List[int],
            int,
            int,
            None,
            id="coerce-dictconfig-interp_with_ref-to-listconfig",
        ),
        param(
            DictConfig({"key": {}}, element_type=Dict[str, int]),
            ListConfig("${interp}", ref_type=List[int]),
            Dict[str, int],
            str,
            int,
            None,
            id="coerce-listconfig-interp_with_ref-to-dictconfig",
        ),
        param(
            DictConfig({"key": []}, element_type=List[int]),
            DictConfig(MISSING),
            List[int],
            int,
            int,
            None,
            id="coerce-dictconfig-missing-to-listconfig",
        ),
        param(
            DictConfig({"key": {}}, element_type=Dict[str, int]),
            ListConfig(MISSING),
            Dict[str, int],
            str,
            int,
            None,
            id="coerce-listconfig-missing-to-dictconfig",
        ),
        param(
            DictConfig({"key": []}, element_type=List[int]),
            DictConfig(MISSING, ref_type=Optional[Dict[str, int]]),
            List[int],
            int,
            int,
            None,
            id="coerce-dictconfig-missing_with_ref-to-listconfig",
        ),
        param(
            DictConfig({"key": {}}, element_type=Dict[str, int]),
            ListConfig(MISSING, ref_type=Optional[List[int]]),
            Dict[str, int],
            str,
            int,
            None,
            id="coerce-listconfig-missing_with_ref-to-dictconfig",
        ),
        param(
            DictConfig({"key": []}, element_type=Optional[List[int]]),
            DictConfig(None),
            Optional[List[int]],
            int,
            int,
            None,
            id="coerce-dictconfig-none-to-listconfig",
        ),
        param(
            DictConfig({"key": {}}, element_type=Optional[Dict[str, int]]),
            ListConfig(None),
            Optional[Dict[str, int]],
            str,
            int,
            None,
            id="coerce-listconfig-none-to-dictconfig",
        ),
        param(
            DictConfig({"key": []}, element_type=Optional[List[int]]),
            DictConfig(None, ref_type=Optional[Dict[str, int]]),
            Optional[List[int]],
            int,
            int,
            None,
            id="coerce-dictconfig-none_with_ref-to-listconfig",
        ),
        param(
            DictConfig({"key": {}}, element_type=Optional[Dict[str, int]]),
            ListConfig(None, ref_type=Optional[List[int]]),
            Optional[Dict[str, int]],
            str,
            int,
            None,
            id="coerce-listconfig-none_with_ref-to-dictconfig",
        ),
    ],
)
def test_dict_assign_to_container_typed_element(
    cfg: DictConfig,
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
    ensure_container: bool,
) -> None:
    cfg = copy.deepcopy(cfg)
    if ensure_container:
        value = _ensure_container(value)

    for key in cfg:
        cfg[key] = value
        check_subnode(cfg, key, value, type_hint, key_type, elt_type, obj_type)

    cfg["_new_key"] = value
    check_subnode(cfg, "_new_key", value, type_hint, key_type, elt_type, obj_type)


@mark.parametrize(
    "dc,value",
    [
        param(DictConfig({"key": 123}, element_type=int), 456, id="int"),
        param(DictConfig({"key": [123]}, element_type=List[int]), [456], id="list"),
        param(
            DictConfig({"key": {"foo": 123}}, element_type=Dict[str, int]),
            {"baz": 456},
            id="dict",
        ),
    ],
)
@mark.parametrize("overwrite_preexisting_key", [True, False])
def test_setitem_valid_element_type(
    dc: DictConfig, value: Any, overwrite_preexisting_key: bool
) -> None:
    dc = copy.deepcopy(dc)
    if not overwrite_preexisting_key:
        del dc["key"]
    dc["key"] = value
    assert dc["key"] == value


@mark.parametrize(
    "cfg, type_hint, key_type, elt_type",
    [
        param(
            DictConfig({"foo": [123], "bar": None}, element_type=Optional[List[int]]),
            Optional[List[int]],
            int,
            int,
            id="assign-to-optional-list",
        ),
        param(
            DictConfig(
                {"foo": {"bar": 456}, "baz": None},
                element_type=Optional[Dict[str, int]],
            ),
            Optional[Dict[str, int]],
            str,
            int,
            id="assign-to-optional-dict",
        ),
    ],
)
@mark.parametrize(
    "value",
    [
        param(None, id="none"),
        param(MISSING, id="missing"),
        param("${interp}", id="interp"),
    ],
)
def test_dict_assign_to_container_typed_element_special(
    cfg: DictConfig,
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
) -> None:
    cfg = copy.deepcopy(cfg)

    for key in cfg:
        cfg[key] = value
        check_subnode(cfg, key, value, type_hint, key_type, elt_type, None)

    cfg["_new_key"] = value
    check_subnode(cfg, "_new_key", value, type_hint, key_type, elt_type, None)


@mark.parametrize(
    "ensure_container",
    [
        param(True, id="container"),
        param(False, id="no_container"),
    ],
)
@mark.parametrize(
    "overwrite_preexisting_key",
    [
        param(True, id="overwrite"),
        param(False, id="no_overwrite"),
    ],
)
@mark.parametrize(
    "dc,value,err_msg",
    [
        param(
            DictConfig({"key": 123}, element_type=int),
            "foo",
            re.escape("Value 'foo' of type 'str' could not be converted to Integer"),
            id="assign-str-to-int",
        ),
        param(
            DictConfig({"key": [123]}, element_type=List[int]),
            "foo",
            re.escape("Invalid value assigned: str is not a ListConfig, list or tuple"),
            id="assign-str-to-list[int]",
        ),
        param(
            DictConfig({"key": {"foo": 123}}, element_type=Dict[str, int]),
            "bar",
            re.escape("Cannot assign str to Dict[str, int]"),
            id="assign-str-to-list[int]",
        ),
        param(
            DictConfig({"key": [123]}, element_type=List[int]),
            None,
            re.escape("field 'key' is not Optional"),
            id="assign-none-to-list[int]",
        ),
        param(
            DictConfig({"key": [123]}, element_type=List[int]),
            456,
            re.escape("Invalid value assigned: int is not a ListConfig, list or tuple"),
            id="assign-int-to-list[int]",
        ),
        param(
            DictConfig({"key": [123]}, element_type=List[int]),
            ["foo"],
            r"(Value 'foo' of type 'str' could not be converted to Integer)"
            + r"|(Value 'foo' \(str\) is incompatible with type hint 'int')",
            id="assign-list[str]-to-list[int]",
        ),
        param(
            DictConfig({"key": [123]}, element_type=List[int]),
            [None],
            r"(Invalid type assigned: NoneType is not a subclass of int)"
            + r"|(Incompatible value 'None' for field of type 'int')",
            id="assign-list[none]-to-list[int]",
        ),
        param(
            DictConfig({"key": [123]}, element_type=List[int]),
            {"baz": 456},
            r"(Invalid value assigned: dict is not a ListConfig, list or tuple)"
            + r"|(Invalid value assigned: DictConfig is not a ListConfig, list or tuple)"
            + r"|(Invalid value assigned: dict does not match type hint typing\.List\[int\])"
            + r"|('DictConfig' is incompatible with type hint 'typing\.List\[int\]')",
            id="assign-dict[str-int]-to-list[int]]",
        ),
        param(
            DictConfig({"key": {"key2": 123}}, element_type=Dict[str, int]),
            {"key2": "foo"},
            r"(Value 'foo' \(str\) is incompatible with type hint 'int')"
            + r"|(Value 'foo' of type 'str' could not be converted to Integer)",
            id="assign_dict[str_str]_to_dict[str_int]",
        ),
        param(
            DictConfig({"key": {"key2": 123}}, element_type=Dict[str, int]),
            [],
            r"(Cannot assign list to Dict\[str, int\])"
            + r"|('ListConfig' is incompatible with type hint 'typing.Dict\[str, int\]')",
            id="assign_list_to_dict[str_int]",
        ),
        param(
            DictConfig({"key": [[123]]}, element_type=List[List[int]]),
            [[456.789]],
            r"(Value 456\.789 \(float\) is incompatible with type hint 'int')"
            + r"|(Value '456\.789' of type 'float' could not be converted to Integer)",
            id="assign-list[list[float]]-to-list[list[int]]",
        ),
        param(
            DictConfig({"key": [[123]]}, element_type=List[List[int]]),
            [[None]],
            r"(Invalid type assigned: NoneType is not a subclass of int)"
            + r"|(Incompatible value 'None' for field of type 'int')",
            id="assign-list[list[none]]-to-list[list[int]]",
        ),
        param(
            DictConfig({"key": [[123]]}, element_type=List[List[int]]),
            [[IntegerNode(None)]],
            r"(Value None \(NoneType\) is incompatible with type hint 'int')"
            + r"|(Incompatible value 'None' for field of type 'int')",
            id="assign-list[list[typed-none]]-to-list[list[int]]",
        ),
        param(
            DictConfig({"key": [[123.456]]}, element_type=List[List[float]]),
            [[IntegerNode(789)]],
            re.escape("Value 789 (int) is incompatible with type hint 'float'"),
            id="assign-list[list[typed-int]]-to-list[list[float]]",
        ),
        param(
            DictConfig(
                {"key": {"foo": {"bar": 123}}}, element_type=Dict[str, Dict[str, int]]
            ),
            {"foo": {"bar": 456.789}},
            r"(Value 456\.789 \(float\) is incompatible with type hint 'int')"
            + r"|(Value '456\.789' of type 'float' could not be converted to Integer)",
            id="assign-dict[str-[dict[str-float]]]-to-dict[str[dict[str-int]]]",
        ),
        param(
            DictConfig(
                {"key": {"foo": {"bar": 123}}}, element_type=Dict[str, Dict[str, int]]
            ),
            {"foo": {"bar2": 456.789}},
            r"(Value 456\.789 \(float\) is incompatible with type hint 'int')"
            + r"|(Value '456\.789' of type 'float' could not be converted to Integer)",
            id="assign-dict[str-[dict[str-float]]]-to-dict[str[dict[str-int]]]-2",
        ),
        param(
            DictConfig(
                {"key": {"foo": {"bar": 123}}}, element_type=Dict[str, Dict[str, int]]
            ),
            {"foo": {123: 456}},
            r"(Key 123 \(int\) is incompatible with \(str\))"
            + r"|(Key 123 \(int\) is incompatible with key type hint 'str')",
            id="assign-dict[str_[dict[int_int]]]-to-dict[str[dict[str_int]]]",
        ),
        param(
            DictConfig(
                {"key": {"foo": {"bar": 123}}}, element_type=Dict[str, Dict[str, int]]
            ),
            {"foo": {456: 789}},
            r"(Key 456 \(int\) is incompatible with \(str\))"
            + r"|(Key 456 \(int\) is incompatible with key type hint 'str')",
            id="assign-dict[str_[dict[int-int]]]-to-dict[str[dict[str_int]]]",
        ),
        param(
            DictConfig(
                {"key": {"foo": {"bar": 123}}}, element_type=Dict[str, Dict[str, int]]
            ),
            {"foo": {456: IntegerNode(None)}},
            r"(Key 456 \(int\) is incompatible with \(str\))"
            + r"|(Key 456 \(int\) is incompatible with key type hint 'str')",
            id="assign-dict[str_[dict[int-typed_none]]]-to-dict[str[dict[str_int]]]",
        ),
        param(
            DictConfig(
                {"key": {"foo": {"bar": 123.456}}},
                element_type=Dict[str, Dict[str, float]],
            ),
            {"foo": {"bar": IntegerNode(789)}},
            re.escape("Value 789 (int) is incompatible with type hint 'float'"),
            id="assign-dict[str-[dict[int_typed-int]]]-to-dict[str[dict[str-float]]]",
        ),
        param(
            DictConfig(
                {"key": {"foo": {"bar": 123.456}}},
                element_type=Dict[str, Dict[str, float]],
            ),
            {"foo": {"bar2": IntegerNode(789)}},
            re.escape("Value 789 (int) is incompatible with type hint 'float'"),
            id="assign-dict[str-[dict[int_typed-int]]]-to-dict[str[dict[str_float]]]-2",
        ),
    ],
)
def test_dict_setitem_invalid_element_type(
    dc: DictConfig,
    value: Any,
    err_msg: str,
    ensure_container: bool,
    overwrite_preexisting_key: bool,
) -> None:
    dc_orig = dc
    dc = copy.deepcopy(dc)

    if ensure_container:
        if isinstance(value, (dict, list)):
            value = _ensure_container(value)
        else:
            return  # skip

    if overwrite_preexisting_key:
        with raises((ValidationError, KeyValidationError), match=err_msg):
            dc["key"] = value
        assert dc == dc_orig
    else:
        del dc["key"]
        with raises((ValidationError, KeyValidationError), match=err_msg):
            dc["key"] = value
        assert dc == {}


@mark.parametrize(
    "lc,index,value,err_msg",
    [
        param(
            ListConfig([123], element_type=int),
            0,
            "foo",
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="assign_str_to_int",
        ),
        param(
            ListConfig([123], element_type=int),
            0,
            None,
            re.escape("[0] is not optional and cannot be assigned None"),
            id="assign_none_to_int",
        ),
        param(
            ListConfig([[123]], element_type=List[int]),
            0,
            "foo",
            "Invalid value assigned: str is not a ListConfig, list or tuple",
            id="assign_str_to_list[int]",
        ),
        param(
            ListConfig([{"key": 123}], element_type=Dict[str, int]),
            0,
            "foo",
            re.escape("Cannot assign str to Dict[str, int]"),
            id="assign_str_to_dict[str, int]",
        ),
        param(
            ListConfig([[123]], element_type=List[int]),
            0,
            None,
            re.escape("[0] is not optional and cannot be assigned None"),
            id="assign_none_to_list[int]",
        ),
        param(
            ListConfig([[123]], element_type=List[int]),
            0,
            456,
            "Invalid value assigned: int is not a ListConfig, list or tuple",
            id="assign_int_to_list[int]",
        ),
        param(
            ListConfig([[123]], element_type=List[int]),
            0,
            ["foo"],
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="assign_list[str]_to_list[int]",
        ),
        param(
            ListConfig([[123]], element_type=List[int]),
            0,
            [None],
            re.escape("Invalid type assigned: NoneType is not a subclass of int"),
            id="assign_list[none]_to_list[int]",
        ),
        param(
            ListConfig([[123]], element_type=List[int]),
            0,
            {"baz": 456},
            "Invalid value assigned: dict",
            id="assign_dict[str,int]_to_list[int]]",
        ),
        param(
            ListConfig([{"key": 123}], element_type=Dict[str, int]),
            0,
            {"key2": "foo"},
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="assign_dict[str,str]_to_dict[str,int]",
        ),
        param(
            ListConfig([{"key2": 123}], element_type=Dict[str, int]),
            0,
            {"key2": "foo"},
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="assign_dict[str,str]_to_dict[str,int]",
        ),
        param(
            ListConfig([], element_type=int),
            None,
            "foo",
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="append_str_to_int",
        ),
        param(
            ListConfig([], element_type=int),
            None,
            None,
            "Invalid type assigned: NoneType is not a subclass of int",
            id="append_none_to_int",
        ),
        param(
            ListConfig([], element_type=List[int]),
            None,
            "foo",
            "Invalid value assigned: str is not a ListConfig, list or tuple",
            id="append_str_to_list[int]",
        ),
        param(
            ListConfig([], element_type=Dict[str, int]),
            None,
            "foo",
            re.escape("Cannot assign str to Dict[str, int]"),
            id="append_str_to_dict[str, int]",
        ),
        param(
            ListConfig([], element_type=List[int]),
            None,
            None,
            re.escape("Invalid type assigned: NoneType is not a subclass of List[int]"),
            id="append_none_to_list[int]",
        ),
        param(
            ListConfig([], element_type=List[int]),
            None,
            456,
            "Invalid value assigned: int is not a ListConfig, list or tuple",
            id="append_int_to_list[int]",
        ),
        param(
            ListConfig([], element_type=List[int]),
            None,
            ["foo"],
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="append_list[str]_to_list[int]",
        ),
        param(
            ListConfig([], element_type=List[int]),
            None,
            [None],
            re.escape("Invalid type assigned: NoneType is not a subclass of int"),
            id="append_list[none]_to_list[int]",
        ),
        param(
            ListConfig([], element_type=List[int]),
            None,
            {"baz": 456},
            "Invalid value assigned: dict",
            id="append_dict[str,int]_to_list[int]]",
        ),
        param(
            ListConfig([], element_type=Dict[str, int]),
            None,
            {"key2": "foo"},
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="append_dict[str,str]_to_dict[str,int]",
        ),
        param(
            ListConfig([], element_type=Dict[str, int]),
            None,
            {"key2": "foo"},
            "Value 'foo' of type 'str' could not be converted to Integer",
            id="append_dict[str,str]_to_dict[str,int]",
        ),
        param(
            ListConfig(
                [{"key2": ConcretePlugin()}], element_type=Dict[str, ConcretePlugin]
            ),
            0,
            {"key": Plugin()},
            "Invalid type assigned: Plugin is not a subclass of ConcretePlugin",
            id="append_dict[str,str]_to_dict[str,int]",
        ),
        param(
            ListConfig([[ConcretePlugin()]], element_type=List[ConcretePlugin]),
            0,
            [Plugin()],
            "Invalid type assigned: Plugin is not a subclass of ConcretePlugin",
            id="append_dict[str,str]_to_dict[str,int]",
        ),
        param(
            ListConfig([], element_type=Dict[str, ConcretePlugin]),
            None,
            {"key": Plugin()},
            "Invalid type assigned: Plugin is not a subclass of ConcretePlugin",
            id="append_dict[str,str]_to_dict[str,int]",
        ),
        param(
            ListConfig([], element_type=List[ConcretePlugin]),
            None,
            [Plugin()],
            "Invalid type assigned: Plugin is not a subclass of ConcretePlugin",
            id="append_dict[str,str]_to_dict[str,int]",
        ),
    ],
)
def test_list_setitem_invalid_element_type(
    lc: ListConfig,
    index: Optional[int],
    value: Any,
    err_msg: str,
) -> None:
    lc_orig = lc
    lc = copy.deepcopy(lc)
    with raises(ValidationError, match=err_msg):
        if index is None:
            lc.append(value)
        else:
            lc[index] = value
    assert lc == lc_orig


@mark.parametrize(
    "dc1, dc2, value, type_hint, key_type, elt_type, obj_type",
    [
        param(
            DictConfig({"key": {"key2": Plugin()}}, element_type=Dict[str, Plugin]),
            DictConfig({"key": {"key2": ConcretePlugin()}}, element_type=Any),
            ConcretePlugin(),
            Plugin,
            Any,
            Any,
            ConcretePlugin,
            id="any-plugin-into-typed-plugin",
        ),
        param(
            DictConfig({"key": {"key2": Plugin()}}, element_type=Any),
            DictConfig(
                {"key": {"key2": ConcretePlugin()}}, element_type=Dict[str, Plugin]
            ),
            ConcretePlugin(),
            Plugin,
            Any,
            Any,
            ConcretePlugin,
            id="typed-plugin-into-any-plugin",
        ),
        param(
            DictConfig({"key": {"key2": Plugin()}}, element_type=Dict[str, Plugin]),
            DictConfig(
                {"key": {"key2": ConcretePlugin()}},
                element_type=Dict[str, ConcretePlugin],
            ),
            ConcretePlugin(),
            Plugin,
            Any,
            Any,
            ConcretePlugin,
            id="typed-concrete-plugin-into-typed-plugin",
        ),
        param(
            DictConfig({"key": {"key2": {}}}),
            DictConfig({"key": {"key2": Plugin()}}, element_type=Dict[str, Plugin]),
            Plugin(),
            Plugin,
            Any,
            Any,
            Plugin,
            id="typed-plugin-into-any",
        ),
    ],
)
def test_merge_nested_dict_promotion(
    dc1: DictConfig,
    dc2: DictConfig,
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
) -> None:
    cfg = OmegaConf.merge(dc1, dc2)
    check_subnode(
        cfg.key,
        key="key2",
        value=value,
        type_hint=type_hint,
        key_type=key_type,
        elt_type=elt_type,
        obj_type=obj_type,
    )


@mark.parametrize(
    "configs, keys, value, type_hint, key_type, elt_type, obj_type",
    [
        param(
            [
                DictConfig({}, element_type=Dict[str, List[int]]),
                DictConfig({"foo": {"bar": "${interp}"}}, element_type=Dict[str, Any]),
            ],
            ["foo", "bar"],
            "${interp}",
            List[int],
            int,
            int,
            None,
            id="merge-interp-into-list",
        ),
        param(
            [
                DictConfig({}, element_type=Dict[str, Optional[List[int]]]),
                DictConfig({"foo": {"bar": None}}, element_type=Dict[str, Any]),
            ],
            ["foo", "bar"],
            None,
            Optional[List[int]],
            int,
            int,
            None,
            id="merge-none-into-list",
        ),
        param(
            [
                DictConfig({}, element_type=Dict[str, Dict[str, int]]),
                DictConfig({"foo": {"bar": "${interp}"}}, element_type=Dict[str, Any]),
            ],
            ["foo", "bar"],
            "${interp}",
            Dict[str, int],
            str,
            int,
            None,
            id="merge-interp-into-dict",
        ),
        param(
            [
                DictConfig({}, element_type=Dict[str, Optional[Dict[str, int]]]),
                DictConfig({"foo": {"bar": None}}, element_type=Dict[str, Any]),
            ],
            ["foo", "bar"],
            None,
            Optional[Dict[str, int]],
            str,
            int,
            None,
            id="merge-none-into-dict",
        ),
    ],
)
def test_merge_nested(
    configs: List[Any],
    keys: List[Any],
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
) -> None:
    """Ensure metadata and contents of container-typed subnode are correct"""
    cfg = OmegaConf.merge(*configs)
    for key in keys[:-1]:
        cfg = cfg._get_node(key)  # type: ignore
    key = keys[-1]
    check_subnode(
        cfg,
        key,
        value,
        type_hint,
        key_type,
        elt_type,
        obj_type,
    )


@mark.parametrize(
    "dc1, dc2, value, type_hint, key_type, elt_type, obj_type",
    [
        param(
            DictConfig({"key": {}}),
            DictConfig({"key": "${interp}"}, element_type=Dict[str, int]),
            "${interp}",
            Dict[str, int],
            str,
            int,
            None,
            id="dict-interp-into-any",
        ),
        param(
            DictConfig({"key": {}}),
            DictConfig({"key": None}, element_type=Optional[Dict[str, int]]),
            None,
            Optional[Dict[str, int]],
            str,
            int,
            None,
            id="none-interp-into-any",
        ),
        param(
            DictConfig({"key": {"foo": 123}}, element_type=Dict[str, Any]),
            DictConfig({"key": {"bar": 456.789}}, element_type=Dict[str, float]),
            {"foo": 123, "bar": 456.789},
            Dict[str, Any],
            str,
            Any,
            dict,
            id="dict[str,float]-into-dict[str,any]",
        ),
        param(
            DictConfig({"key": {}}, element_type=Dict[str, int]),
            DictConfig({"key": "${interp}"}),
            "${interp}",
            Dict[str, int],
            str,
            int,
            None,
            id="interp-into-dict",
        ),
        param(
            DictConfig({"key": []}),
            DictConfig({"key": "${interp}"}, element_type=List[int]),
            "${interp}",
            Any,
            int,
            Any,
            None,
            id="list-interp-into-any",
        ),
        param(
            DictConfig({"key": []}, element_type=List[int]),
            DictConfig({"key": "${interp}"}),
            "${interp}",
            List[int],
            int,
            int,
            None,
            id="any-interp-into-list-int",
        ),
        param(
            DictConfig({"key": []}, element_type=List[float]),
            DictConfig({"key": ["${interp}"]}, element_type=List[int]),
            ["${interp}"],
            List[float],
            int,
            float,
            list,
            id="any-interp_list-into-list-list-int",
        ),
    ],
)
def test_merge_interpolation_with_container_type(
    dc1: DictConfig,
    dc2: DictConfig,
    value: Any,
    type_hint: Any,
    key_type: Any,
    elt_type: Any,
    obj_type: Any,
) -> None:
    cfg = OmegaConf.merge(dc1, dc2)
    check_subnode(
        cfg,
        key="key",
        value=value,
        type_hint=type_hint,
        key_type=key_type,
        elt_type=elt_type,
        obj_type=obj_type,
    )


def test_merge_nested_list_promotion() -> None:
    dc1 = DictConfig({"key": [Plugin]}, element_type=List[Plugin])
    dc2 = DictConfig({"key": [ConcretePlugin]})
    cfg = OmegaConf.merge(dc1, dc2)
    check_subnode(
        cfg.key,
        key=0,
        value=ConcretePlugin(),
        type_hint=Plugin,
        key_type=Any,
        elt_type=Any,
        obj_type=ConcretePlugin,
    )


@mark.parametrize(
    "configs, err_msg",
    [
        param(
            [DictConfig({}, element_type=int), {"foo": "abc"}],
            "Value 'abc' of type 'str' could not be converted to Integer",
        ),
        param(
            [DictConfig({}, element_type=Dict[str, int]), {"foo": 123}],
            "Value 123 (int) is incompatible with type hint 'typing.Dict[str, int]'",
            id="merge-int-into-dict",
        ),
        param(
            [
                DictConfig({}, element_type=Dict[str, Dict[str, int]]),
                DictConfig(
                    {"foo": {"bar": None}}, element_type=Dict[str, Optional[int]]
                ),
            ],
            "field 'foo.bar' is not Optional",
            id="merge-none_typed-into-int",
        ),
    ],
)
def test_merge_bad_element_type(configs: Any, err_msg: Any) -> None:
    with raises(
        ValidationError,
        match=re.escape(err_msg),
    ):
        OmegaConf.merge(*configs)



import pathlib
import platform
from pathlib import Path
from typing import Any, Union

from pytest import mark, param, raises

from omegaconf import (
    MISSING,
    BooleanNode,
    BytesNode,
    DictConfig,
    EnumNode,
    FloatNode,
    IntegerNode,
    ListConfig,
    MissingMandatoryValue,
    OmegaConf,
    PathNode,
    StringNode,
    UnionNode,
)
from omegaconf._utils import _is_none, nullcontext
from omegaconf.errors import (
    ConfigKeyError,
    InterpolationKeyError,
    InterpolationToMissingValueError,
    UnsupportedInterpolationType,
)
from tests import Color, ConcretePlugin, IllegalType, StructuredWithMissing


@mark.parametrize(
    "cfg, key, expected_is_missing, expectation",
    [
        ({}, "foo", False, raises(ConfigKeyError)),
        ({"foo": True}, "foo", False, nullcontext()),
        ({"foo": "${no_such_key}"}, "foo", False, raises(InterpolationKeyError)),
        ({"foo": MISSING}, "foo", True, raises(MissingMandatoryValue)),
        param(
            {"foo": "${bar}", "bar": DictConfig(content=MISSING)},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
            id="missing_interpolated_dict",
        ),
        param(
            {"foo": ListConfig(content="???")},
            "foo",
            True,
            raises(MissingMandatoryValue),
            id="missing_list",
        ),
        param(
            {"foo": DictConfig(content="???")},
            "foo",
            True,
            raises(MissingMandatoryValue),
            id="missing_dict",
        ),
        param(
            {"foo": ListConfig(content="${missing}"), "missing": "???"},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
            id="missing_list_interpolation",
        ),
        param(
            {"foo": DictConfig(content="${missing}"), "missing": "???"},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
            id="missing_dict_interpolation",
        ),
        (
            {"foo": "${bar}", "bar": MISSING},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            {"foo": "foo_${bar}", "bar": MISSING},
            "foo",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            {"foo": "${unknown_resolver:foo}"},
            "foo",
            False,
            raises(UnsupportedInterpolationType),
        ),
        ({"foo": StringNode(value="???")}, "foo", True, raises(MissingMandatoryValue)),
        (
            {"foo": StringNode(value="???"), "inter": "${foo}"},
            "inter",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (StructuredWithMissing, "num", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_num", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "dict", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_dict", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "list", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_list", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "user", True, raises(MissingMandatoryValue)),
        (StructuredWithMissing, "opt_user", True, raises(MissingMandatoryValue)),
        (
            StructuredWithMissing,
            "inter_user",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            StructuredWithMissing,
            "inter_opt_user",
            False,
            raises(InterpolationToMissingValueError),
        ),
        (
            StructuredWithMissing,
            "inter_num",
            False,
            raises(InterpolationToMissingValueError),
        ),
    ],
)
def test_is_missing(
    cfg: Any, key: str, expected_is_missing: bool, expectation: Any
) -> None:
    cfg = OmegaConf.create(cfg)
    with expectation:
        cfg[key]

    assert OmegaConf.is_missing(cfg, key) == expected_is_missing
    OmegaConf.set_struct(cfg, True)
    assert OmegaConf.is_missing(cfg, key) == expected_is_missing
    OmegaConf.set_readonly(cfg, True)
    assert OmegaConf.is_missing(cfg, key) == expected_is_missing


def test_is_missing_resets() -> None:
    cfg = OmegaConf.structured(StructuredWithMissing)
    assert OmegaConf.is_missing(cfg, "dict")
    cfg.dict = {}
    assert not OmegaConf.is_missing(cfg, "dict")

    assert OmegaConf.is_missing(cfg, "list")
    cfg.list = [1, 2, 3]
    assert not OmegaConf.is_missing(cfg, "list")
    cfg.list = "???"
    assert OmegaConf.is_missing(cfg, "list")


@mark.parametrize(
    "cfg, expected",
    [
        (None, False),
        ({}, False),
        ([], False),
        ("aa", False),
        (10, False),
        (True, False),
        (bool, False),
        (Path, False),
        (Path("hello.txt"), False),
        (StringNode("foo"), False),
        (ConcretePlugin, False),
        (ConcretePlugin(), False),
        (OmegaConf.create({}), True),
        (OmegaConf.create([]), True),
        (OmegaConf.structured(ConcretePlugin), True),
        (OmegaConf.structured(ConcretePlugin()), True),
    ],
)
def test_is_config(cfg: Any, expected: bool) -> None:
    assert OmegaConf.is_config(cfg) == expected


@mark.parametrize(
    "cfg, expected",
    [
        (None, False),
        ({}, False),
        ([], False),
        ("aa", False),
        (10, False),
        (True, False),
        (bool, False),
        (Path, False),
        (Path("hello.txt"), False),
        (StringNode("foo"), False),
        (ConcretePlugin, False),
        (ConcretePlugin(), False),
        (OmegaConf.create({}), False),
        (OmegaConf.create([]), True),
        (OmegaConf.structured(ConcretePlugin), False),
        (OmegaConf.structured(ConcretePlugin()), False),
    ],
)
def test_is_list(cfg: Any, expected: bool) -> None:
    assert OmegaConf.is_list(cfg) == expected


@mark.parametrize(
    "cfg, expected",
    [
        (None, False),
        ({}, False),
        ([], False),
        ("aa", False),
        (10, False),
        (True, False),
        (bool, False),
        (Path, False),
        (Path("hello.txt"), False),
        (StringNode("foo"), False),
        (ConcretePlugin, False),
        (ConcretePlugin(), False),
        (OmegaConf.create({}), True),
        (OmegaConf.create([]), False),
        (OmegaConf.structured(ConcretePlugin), True),
        (OmegaConf.structured(ConcretePlugin()), True),
    ],
)
def test_is_dict(cfg: Any, expected: bool) -> None:
    assert OmegaConf.is_dict(cfg) == expected


@mark.parametrize("is_none", [True, False])
@mark.parametrize(
    "fac",
    [
        (lambda none: StringNode(value="foo" if not none else None, is_optional=True)),
        (lambda none: IntegerNode(value=10 if not none else None, is_optional=True)),
        (lambda none: FloatNode(value=10.0 if not none else None, is_optional=True)),
        (lambda none: BooleanNode(value=True if not none else None, is_optional=True)),
        (lambda none: BytesNode(value=b"123" if not none else None, is_optional=True)),
        (
            lambda none: PathNode(
                value=Path("hello.txt") if not none else None, is_optional=True
            )
        ),
        (
            lambda none: EnumNode(
                enum_type=Color,
                value=Color.RED if not none else None,
                is_optional=True,
            )
        ),
        (
            lambda none: ListConfig(
                content=[1, 2, 3] if not none else None, is_optional=True
            )
        ),
        (
            lambda none: DictConfig(
                content={"foo": "bar"} if not none else None, is_optional=True
            )
        ),
        (
            lambda none: DictConfig(
                ref_type=ConcretePlugin,
                content=ConcretePlugin() if not none else None,
                is_optional=True,
            )
        ),
    ],
)
def test_is_none(fac: Any, is_none: bool) -> None:
    obj = fac(is_none)
    assert obj._is_none() == is_none

    cfg = OmegaConf.create({"node": obj})
    assert (cfg.node is None) == is_none


@mark.parametrize(
    ("cfg", "key", "is_none"),
    [
        ({"foo": "${none}", "none": None}, "foo", True),
        ({"foo": "${missing_node}"}, "foo", False),
        ({"foo": "${missing_resolver:}"}, "foo", False),
        ({"foo": "${oc.decode:'[1, '}"}, "foo", False),
    ],
)
def test_is_none_interpolation(cfg: Any, key: str, is_none: bool) -> None:
    cfg = OmegaConf.create(cfg)
    check = _is_none(
        cfg._get_node(key), resolve=True, throw_on_resolution_failure=False
    )
    assert check == is_none


@mark.parametrize(
    "fac",
    [
        (
            lambda inter: StringNode(
                value="foo" if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: IntegerNode(
                value=10 if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: FloatNode(
                value=10 if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: BooleanNode(
                value=True if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: BytesNode(
                value=b"123" if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: PathNode(
                value="hello.txt" if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: EnumNode(
                enum_type=Color,
                value=Color.RED if inter is None else inter,
                is_optional=True,
            )
        ),
        (
            lambda inter: ListConfig(
                content=[1, 2, 3] if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: DictConfig(
                content={"foo": "bar"} if inter is None else inter, is_optional=True
            )
        ),
        (
            lambda inter: DictConfig(
                ref_type=ConcretePlugin,
                content=ConcretePlugin() if inter is None else inter,
                is_optional=True,
            )
        ),
    ],
    ids=[
        "StringNode",
        "IntegerNode",
        "FloatNode",
        "BooleanNode",
        "BytesNode",
        "PathNode",
        "EnumNode",
        "ListConfig",
        "DictConfig",
        "ConcretePlugin",
    ],
)
def test_is_interpolation(fac: Any) -> Any:
    obj = fac(inter=None)
    assert not OmegaConf.is_interpolation(obj)
    cfg = OmegaConf.create({"node": obj})
    assert not OmegaConf.is_interpolation(cfg, "node")

    assert not OmegaConf.is_interpolation(cfg, "missing")

    for inter in ["${foo}", "http://${url}", "${resolver:value}"]:
        obj = fac(inter=inter)
        assert OmegaConf.is_interpolation(obj)
        cfg = OmegaConf.create({"node": obj})
        assert OmegaConf.is_interpolation(cfg, "node")


@mark.parametrize(
    "cfg, type_",
    [
        ({"foo": 10}, int),
        ({"foo": 10.0}, float),
        ({"foo": True}, bool),
        ({"foo": b"123"}, bytes),
        ({"foo": UnionNode(10.0, Union[float, bytes])}, float),
        ({"foo": UnionNode(None, Union[float, bytes])}, type(None)),
        ({"foo": FloatNode(10.0)}, float),
        ({"foo": FloatNode(None)}, type(None)),
        (
            {"foo": Path("hello.txt")},
            pathlib.WindowsPath
            if platform.system() == "Windows"
            else pathlib.PosixPath,
        ),
        ({"foo": "bar"}, str),
        ({"foo": None}, type(None)),
        ({"foo": ConcretePlugin()}, ConcretePlugin),
        ({"foo": ConcretePlugin}, ConcretePlugin),
        ({"foo": {}}, dict),
        ({"foo": OmegaConf.create()}, dict),
        ({"foo": []}, list),
        ({"foo": OmegaConf.create([])}, list),
    ],
)
def test_get_type(cfg: Any, type_: Any) -> None:
    cfg = OmegaConf.create(cfg)
    assert OmegaConf.get_type(cfg, "foo") == type_


@mark.parametrize(
    "obj, type_",
    [
        (10, int),
        (10.0, float),
        (True, bool),
        (b"123", bytes),
        (
            Path("hello.txt"),
            pathlib.WindowsPath
            if platform.system() == "Windows"
            else pathlib.PosixPath,
        ),
        ("foo", str),
        (DictConfig(content={}), dict),
        (ListConfig(content=[]), list),
        (IllegalType, IllegalType),
        (IllegalType(), IllegalType),
    ],
)
def test_get_type_on_raw(obj: Any, type_: Any) -> None:
    assert OmegaConf.get_type(obj) == type_


def test_is_issubclass() -> None:
    cfg = OmegaConf.structured(ConcretePlugin)
    t = OmegaConf.get_type(cfg)
    assert t is not None and issubclass(t, ConcretePlugin)


@mark.parametrize(
    ("cfg", "expected"),
    [
        # dict
        param(OmegaConf.create(), {}, id="dict"),
        param(OmegaConf.create({"a": 10, "b": "${a}"}), {"a": 10, "b": 10}, id="dict"),
        param(
            OmegaConf.create({"a": 10, "b": {"a": "${a}"}}),
            {"a": 10, "b": {"a": 10}},
            id="dict",
        ),
        param(
            OmegaConf.create({"a": "${b.a}", "b": {"a": 10}}),
            {"a": 10, "b": {"a": 10}},
            id="dict",
        ),
        param(OmegaConf.create({"a": "???"}), {"a": "???"}, id="dict:missing"),
        param(
            OmegaConf.create({"a": "???", "b": "${a}"}),
            {"a": "???", "b": "???"},
            id="dict:missing",
        ),
        param(
            OmegaConf.create({"a": 10, "b": "a_${a}"}),
            {"a": 10, "b": "a_10"},
            id="dict:str_inter",
        ),
        # This seems like a reasonable resolution for a string interpolation pointing to a missing node:
        param(
            OmegaConf.create({"a": "???", "b": "a_${a}"}),
            {"a": "???", "b": "???"},
            id="dict:str_inter_missing",
        ),
        param(
            DictConfig("${a}", parent=OmegaConf.create({"a": {"b": 10}})),
            {"b": 10},
            id="inter_dict",
        ),
        param(
            DictConfig("${a}", parent=OmegaConf.create({"a": "???"})),
            "???",
            id="inter_dict_to_missing",
        ),
        param(
            OmegaConf.create({"x": "${y}", "y": {"z": "${foo}"}, "foo": 0}),
            {"x": {"z": 0}, "y": {"z": 0}, "foo": 0},
            id="dict_nested_interpolation",
        ),
        param(DictConfig(None), None, id="none_dict"),
        # comparing to ??? because to_container returns it.
        param(DictConfig("???"), "???", id="missing_dict"),
        # lists
        param(OmegaConf.create([]), [], id="list"),
        param(OmegaConf.create([10, "${0}"]), [10, 10], id="list"),
        param(OmegaConf.create(["???"]), ["???"], id="list:missing"),
        param(OmegaConf.create(["${1}", "???"]), ["???", "???"], id="list:missing"),
        param(
            ListConfig("${a}", parent=OmegaConf.create({"a": [1, 2]})),
            [1, 2],
            id="inter_list",
        ),
        param(
            ListConfig("${a}", parent=OmegaConf.create({"a": "???"})),
            "???",
            id="inter_list_to_missing",
        ),
        param(ListConfig(None), None, id="none_list"),
        # comparing to ??? because to_container returns it.
        param(ListConfig("???"), "???", id="missing_list"),
        # Tests for cases where an AnyNode with interpolation is promoted to a DictConfig/ListConfig node
        param(
            OmegaConf.create({"a": "${z}", "z": {"y": 1}}),
            {"a": {"y": 1}, "z": {"y": 1}},
            id="any_in_dict_to_dict",
        ),
        param(
            OmegaConf.create({"a": "${z}", "z": [1, 2]}),
            {"a": [1, 2], "z": [1, 2]},
            id="any_in_dict_to_list",
        ),
        param(
            OmegaConf.create(["${1}", {"z": {"y": 1}}]),
            [{"z": {"y": 1}}, {"z": {"y": 1}}],
            id="any_in_list_to_dict",
        ),
        param(
            OmegaConf.create(["${1}", [1, 2]]),
            [[1, 2], [1, 2]],
            id="any_in_list_to_list",
        ),
    ],
)
def test_resolve(cfg: Any, expected: Any) -> None:
    OmegaConf.resolve(cfg)
    # convert output to plain dict to avoid smart OmegaConf eq logic
    resolved_plain = OmegaConf.to_container(cfg, resolve=False)
    assert resolved_plain == expected


def test_resolve_invalid_input() -> None:
    with raises(ValueError):
        OmegaConf.resolve("aaa")  # type: ignore


@mark.parametrize(
    "cfg, expected",
    [
        # dict:
        ({"a": 10, "b": {"c": "???", "d": "..."}}, {"b.c"}),
        (
            {
                "a": "???",
                "b": {
                    "foo": "bar",
                    "bar": "???",
                    "more": {"missing": "???", "available": "yes"},
                },
                Color.GREEN: {"tint": "???", "default": Color.BLUE},
            },
            {"a", "b.bar", "b.more.missing", "GREEN.tint"},
        ),
        (
            {"a": "a", "b": {"foo": "bar", "bar": "foo"}},
            set(),
        ),
        (
            {"foo": "bar", "bar": "???", "more": {"foo": "???", "bar": "foo"}},
            {"bar", "more.foo"},
        ),
        # list:
        (["???", "foo", "bar", "???", 77], {"[0]", "[3]"}),
        (["", "foo", "bar"], set()),
        (["foo", "bar", "???"], {"[2]"}),
        (["foo", "???", ["???", "bar"]], {"[1]", "[2][0]"}),
        # mixing:
        (
            [
                "???",
                "foo",
                {
                    "a": True,
                    "b": "???",
                    "c": ["???", None],
                    "d": {"e": "???", "f": "fff", "g": [True, "???"]},
                },
                "???",
                77,
            ],
            {"[0]", "[2].b", "[2].c[0]", "[2].d.e", "[2].d.g[1]", "[3]"},
        ),
        (
            {
                "list": [
                    0,
                    DictConfig({"foo": "???", "bar": None}),
                    "???",
                    ["???", 3, False],
                ],
                "x": "y",
                "y": "???",
            },
            {"list[1].foo", "list[2]", "list[3][0]", "y"},
        ),
        ({Color.RED: ["???", {"missing": "???"}]}, {"RED[0]", "RED[1].missing"}),
        # with DictConfig and ListConfig:
        (
            DictConfig(
                {
                    "foo": "???",
                    "list": ["???", 1],
                    "bar": {"missing": "???", "more": "yes"},
                }
            ),
            {"foo", "list[0]", "bar.missing"},
        ),
        (
            ListConfig(
                ["???", "yes", "???", [0, 1, "???"], {"missing": "???", "more": ""}],
            ),
            {"[0]", "[2]", "[3][2]", "[4].missing"},
        ),
    ],
)
def test_missing_keys(cfg: Any, expected: Any) -> None:
    assert OmegaConf.missing_keys(cfg) == expected


@mark.parametrize("cfg", [float, int])
def test_missing_keys_invalid_input(cfg: Any) -> None:
    with raises(ValueError):
        OmegaConf.missing_keys(cfg)


@mark.parametrize(
    ("register_resolver_params", "name", "expected"),
    [
        param(
            dict(
                name="iamnew",
                resolver=lambda x: str(x).lower(),
                use_cache=False,
                replace=False,
            ),
            "iamnew",
            dict(pre_clear=True, result=True),
            id="remove-new-custom-resolver",
        ),
        param(
            dict(),
            "oc.env",
            dict(pre_clear=True, result=True),
            id="remove-default-resolver",
        ),
        param(
            dict(),
            "idonotexist",
            dict(pre_clear=False, result=False),
            id="remove-nonexistent-resolver",
        ),
    ],
)
def test_clear_resolver(
    restore_resolvers: Any, register_resolver_params: Any, name: str, expected: Any
) -> None:
    if register_resolver_params:
        OmegaConf.register_new_resolver(**register_resolver_params)
    assert expected["pre_clear"] == OmegaConf.has_resolver(name)

    assert OmegaConf.clear_resolver(name) == expected["result"]
    assert not OmegaConf.has_resolver(name)



import re
import sys
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import attr
from pytest import mark, param, raises

from omegaconf import DictConfig, ListConfig, Node, OmegaConf, UnionNode, _utils
from omegaconf._utils import (
    Marker,
    NoneType,
    _ensure_container,
    _get_value,
    _is_optional,
    _resolve_optional,
    get_dict_key_value_types,
    get_list_element_type,
    is_dict_annotation,
    is_list_annotation,
    is_primitive_dict,
    is_primitive_list,
    is_supported_union_annotation,
    is_tuple_annotation,
    is_union_annotation,
    nullcontext,
    split_key,
)
from omegaconf.errors import UnsupportedValueType, ValidationError
from omegaconf.nodes import (
    AnyNode,
    BooleanNode,
    BytesNode,
    EnumNode,
    FloatNode,
    IntegerNode,
    PathNode,
    StringNode,
)
from omegaconf.omegaconf import _node_wrap
from tests import (
    Color,
    ConcretePlugin,
    Dataframe,
    DictSubclass,
    IllegalType,
    ListSubclass,
    Plugin,
    Shape,
    Str2Int,
    UnionAnnotations,
    User,
)


class DummyEnum(Enum):
    FOO = 1


@mark.parametrize("is_optional", [True, False])
@mark.parametrize(
    "ref_type, value, expected_type",
    [
        (Any, 10, AnyNode),
        (DummyEnum, DummyEnum.FOO, EnumNode),
        (int, 42, IntegerNode),
        (bytes, b"\xf0\xf1\xf2", BytesNode),
        (Path, Path("hello.txt"), PathNode),
        (float, 3.1415, FloatNode),
        (bool, True, BooleanNode),
        (str, "foo", StringNode),
    ],
)
def test_node_wrap(
    ref_type: type, is_optional: bool, value: Any, expected_type: Any
) -> None:
    from omegaconf.omegaconf import _node_wrap

    ret = _node_wrap(
        ref_type=ref_type,
        value=value,
        is_optional=is_optional,
        parent=None,
        key=None,
    )
    assert ret._metadata.ref_type == ref_type
    assert type(ret) == expected_type
    assert ret == value

    if is_optional:
        ret = _node_wrap(
            ref_type=ref_type,
            value=None,
            is_optional=is_optional,
            parent=None,
            key=None,
        )
        assert type(ret) == expected_type
        # noinspection PyComparisonWithNone
        assert ret == None  # noqa E711


@mark.parametrize(
    "target_type, value, expected",
    [
        # Any
        param(Any, "foo", AnyNode("foo"), id="any"),
        param(Any, b"binary", AnyNode(b"binary"), id="any"),
        param(Any, Path("hello.txt"), AnyNode(Path("hello.txt")), id="any"),
        param(Any, True, AnyNode(True), id="any"),
        param(Any, 1, AnyNode(1), id="any"),
        param(Any, 1.0, AnyNode(1.0), id="any"),
        param(Any, b"123", AnyNode(b"123"), id="any"),
        param(Any, Color.RED, AnyNode(Color.RED), id="any"),
        param(Any, {}, DictConfig(content={}), id="any_as_dict"),
        param(Any, [], ListConfig(content=[]), id="any_as_list"),
        # int
        param(int, "foo", ValidationError, id="int"),
        param(int, b"binary", ValidationError, id="int"),
        param(int, Path("hello.txt"), ValidationError, id="int"),
        param(int, True, ValidationError, id="int"),
        param(int, 1, IntegerNode(1), id="int"),
        param(int, 1.0, ValidationError, id="int"),
        param(int, Color.RED, ValidationError, id="int"),
        param(int, b"123", ValidationError, id="int"),
        # float
        param(float, "foo", ValidationError, id="float"),
        param(float, b"binary", ValidationError, id="float"),
        param(float, Path("hello.txt"), ValidationError, id="float"),
        param(float, True, ValidationError, id="float"),
        param(float, 1, FloatNode(1), id="float"),
        param(float, 1.0, FloatNode(1.0), id="float"),
        param(float, Color.RED, ValidationError, id="float"),
        param(float, b"123", ValidationError, id="float"),
        # bool
        param(bool, "foo", ValidationError, id="bool"),
        param(bool, b"binary", ValidationError, id="bool"),
        param(bool, Path("hello.txt"), ValidationError, id="bool"),
        param(bool, True, BooleanNode(True), id="bool"),
        param(bool, 1, BooleanNode(True), id="bool"),
        param(bool, 0, BooleanNode(False), id="bool"),
        param(bool, 1.0, ValidationError, id="bool"),
        param(bool, Color.RED, ValidationError, id="bool"),
        param(bool, "true", BooleanNode(True), id="bool"),
        param(bool, "false", BooleanNode(False), id="bool"),
        param(bool, "on", BooleanNode(True), id="bool"),
        param(bool, "off", BooleanNode(False), id="bool"),
        param(bool, b"123", ValidationError, id="bool"),
        # str
        param(str, "foo", StringNode("foo"), id="str"),
        param(str, b"binary", ValidationError, id="str"),
        param(str, Path("hello.txt"), StringNode("hello.txt"), id="str"),
        param(str, True, StringNode("True"), id="str"),
        param(str, 1, StringNode("1"), id="str"),
        param(str, 1.0, StringNode("1.0"), id="str"),
        param(str, Color.RED, StringNode("Color.RED"), id="str"),
        # bytes
        param(bytes, "foo", ValidationError, id="bytes"),
        param(bytes, b"binary", BytesNode(b"binary"), id="bytes"),
        param(bytes, Path("hello.txt"), ValidationError, id="bytes"),
        param(bytes, True, ValidationError, id="bytes"),
        param(bytes, 1, ValidationError, id="bytes"),
        param(bytes, 1.0, ValidationError, id="bytes"),
        param(bytes, Color.RED, ValidationError, id="bytes"),
        # Path
        param(Path, "foo", PathNode("foo"), id="path"),
        param(Path, b"binary", ValidationError, id="path"),
        param(Path, Path("hello.txt"), PathNode("hello.txt"), id="path"),
        param(Path, True, ValidationError, id="path"),
        param(Path, 1, ValidationError, id="path"),
        param(Path, 1.0, ValidationError, id="path"),
        param(Path, Color.RED, ValidationError, id="path"),
        # Color
        param(Color, "foo", ValidationError, id="Color"),
        param(Color, True, ValidationError, id="Color"),
        param(Color, 1, EnumNode(enum_type=Color, value=Color.RED), id="Color"),
        param(Color, 1.0, ValidationError, id="Color"),
        param(Color, Color.RED, EnumNode(enum_type=Color, value=Color.RED), id="Color"),
        param(Color, "RED", EnumNode(enum_type=Color, value=Color.RED), id="Color"),
        param(
            Color, "Color.RED", EnumNode(enum_type=Color, value=Color.RED), id="Color"
        ),
        param(Color, b"123", ValidationError, id="Color"),
        param(Color, Path("hello.txt"), ValidationError, id="Color"),
        param(
            Color, "Color.RED", EnumNode(enum_type=Color, value=Color.RED), id="Color"
        ),
        # bad type
        param(IllegalType, "nope", ValidationError, id="bad_type"),
        param(IllegalType, [1, 2, 3], ValidationError, id="list_bad_type"),
        param(IllegalType, {"foo": "bar"}, ValidationError, id="dict_bad_type"),
        # DictConfig
        param(
            Dict[Any, Any],
            {"foo": "bar"},
            DictConfig(content={"foo": "bar"}),
            id="DictConfig",
        ),
        param(List[Any], {"foo": "bar"}, ValidationError, id="dict_to_list"),
        param(List[int], {"foo": "bar"}, ValidationError, id="dict_to_list[int]"),
        param(
            Any,
            {"foo": "bar"},
            DictConfig(content={"foo": "bar"}),
            id="dict_to_any",
        ),
        param(Plugin, {"foo": "bar"}, ValidationError, id="dict_to_plugin"),
        # Structured Config
        param(Plugin, Plugin(), DictConfig(content=Plugin()), id="DictConfig[Plugin]"),
        param(Any, Plugin(), DictConfig(content=Plugin()), id="plugin_to_any"),
        param(
            Dict[str, int],
            Plugin(),
            ValidationError,
            id="plugin_to_dict[str, int]",
        ),
        param(List[Any], Plugin(), ValidationError, id="plugin_to_list"),
        param(List[int], Plugin(), ValidationError, id="plugin_to_list[int]"),
        # ListConfig
        param(List[Any], [1, 2, 3], ListConfig(content=[1, 2, 3]), id="ListConfig"),
        param(Dict[Any, Any], [1, 2, 3], ValidationError, id="list_to_dict"),
        param(Dict[str, int], [1, 2, 3], ValidationError, id="list_to_dict[str-int]"),
        param(Any, [1, 2, 3], ListConfig(content=[1, 2, 3]), id="list_to_any"),
    ],
)
def test_node_wrap2(target_type: Any, value: Any, expected: Any) -> None:
    from omegaconf.omegaconf import _maybe_wrap

    if isinstance(expected, Node):
        res = _node_wrap(
            ref_type=target_type, key="foo", value=value, is_optional=False, parent=None
        )
        assert type(res) == type(expected)
        assert res == expected
        assert res._key() == "foo"
    else:
        with raises(expected):
            _maybe_wrap(
                ref_type=target_type,
                key=None,
                value=value,
                is_optional=False,
                parent=None,
            )


def test_node_wrap_illegal_type() -> None:
    class UserClass:
        pass

    from omegaconf.omegaconf import _node_wrap

    with raises(ValidationError):
        _node_wrap(
            ref_type=UserClass,
            value=UserClass(),
            is_optional=False,
            parent=None,
            key=None,
        )


class _TestEnum(Enum):
    A = 1
    B = 2


@dataclass
class _TestDataclass:
    x: int = 10
    s: str = "foo"
    b: bool = True
    p: Path = Path("hello.txt")
    d: bytes = b"123"
    f: float = 3.14
    e: _TestEnum = _TestEnum.A
    list1: List[int] = field(default_factory=list)
    dict1: Dict[str, int] = field(default_factory=dict)
    init_false: str = field(init=False, default="foo")


@attr.s(auto_attribs=True)
class _TestAttrsClass:
    x: int = 10
    s: str = "foo"
    b: bool = True
    p: Path = Path("hello.txt")
    d: bytes = b"123"
    f: float = 3.14
    e: _TestEnum = _TestEnum.A
    list1: List[int] = []
    dict1: Dict[str, int] = {}
    init_false: str = attr.field(init=False, default="foo")


@dataclass
class _TestDataclassIllegalValue:
    x: Any = IllegalType()


@attr.s(auto_attribs=True)
class _TestAttrllegalValue:
    x: Any = IllegalType()


class _TestUserClass:
    pass


@mark.parametrize(
    "type_, expected",
    [
        (int, True),
        (float, True),
        (bool, True),
        (str, True),
        (bytes, True),
        (Path, True),
        (Any, True),
        (_TestEnum, True),
        (_TestUserClass, False),
        # Nesting structured configs in contain
        (_TestAttrsClass, True),
        (_TestDataclass, True),
        # container annotations
        (List[int], True),
        (Dict[str, int], True),
        # optional and union
        (Optional[int], True),
        (Union[int, str], True),
        (Union[int, List[str]], False),
        (Union[int, Dict[int, str]], False),
        (Union[int, _TestEnum], True),
        (Union[int, _TestAttrsClass], False),
        (Union[int, _TestDataclass], False),
        (Union[int, _TestUserClass], False),
    ],
)
def test_is_valid_value_annotation(type_: type, expected: bool) -> None:
    from omegaconf._utils import is_valid_value_annotation

    assert is_valid_value_annotation(type_) == expected


class TestGetStructuredConfigInfo:
    @mark.parametrize(
        "test_cls_or_obj",
        [_TestDataclass, _TestDataclass(), _TestAttrsClass, _TestAttrsClass()],
    )
    def test_get_structured_config_data(self, test_cls_or_obj: Any) -> None:
        d = _utils.get_structured_config_data(test_cls_or_obj)
        assert d["x"] == 10
        assert d["s"] == "foo"
        assert d["b"] == bool(True)
        assert d["p"] == Path("hello.txt")
        assert d["d"] == b"123"
        assert d["f"] == 3.14
        assert d["e"] == _TestEnum.A
        assert d["list1"] == []
        assert d["dict1"] == {}

    def test_get_structured_config_data_throws_ValueError(self) -> None:
        with raises(ValueError):
            _utils.get_structured_config_data("invalid")

    @mark.parametrize(
        "test_cls_or_obj",
        [_TestDataclass, _TestDataclass(), _TestAttrsClass, _TestAttrsClass()],
    )
    def test_get_structured_config_field_names(self, test_cls_or_obj: Any) -> None:
        field_names = _utils.get_structured_config_init_field_names(test_cls_or_obj)
        assert field_names == ["x", "s", "b", "p", "d", "f", "e", "list1", "dict1"]

    def test_get_structured_config_field_names_throws_ValueError(self) -> None:
        with raises(ValueError):
            _utils.get_structured_config_init_field_names("invalid")


@mark.parametrize(
    "test_cls",
    [
        _TestDataclassIllegalValue,
        _TestAttrllegalValue,
    ],
)
def test_get_structured_config_data_illegal_value(test_cls: Any) -> None:
    with raises(UnsupportedValueType):
        _utils.get_structured_config_data(test_cls, allow_objects=None)

    with raises(UnsupportedValueType):
        _utils.get_structured_config_data(test_cls, allow_objects=False)

    d = _utils.get_structured_config_data(test_cls, allow_objects=True)
    assert d["x"] == IllegalType()


def test_is_dataclass(mocker: Any) -> None:
    @dataclass
    class Foo:
        pass

    assert _utils.is_dataclass(Foo)
    assert _utils.is_dataclass(Foo())
    assert not _utils.is_dataclass(10)

    mocker.patch("omegaconf._utils.dataclasses", None)
    assert not _utils.is_dataclass(10)


def test_is_attr_class(mocker: Any) -> None:
    @attr.s
    class Foo:
        pass

    assert _utils.is_attr_class(Foo)
    assert _utils.is_attr_class(Foo())

    assert not _utils.is_attr_class(10)
    mocker.patch("omegaconf._utils.attr", None)
    assert not _utils.is_attr_class(10)


def test_is_structured_config_frozen_with_invalid_obj() -> None:
    assert not _utils.is_structured_config_frozen(10)


@dataclass
class Dataclass:
    pass


@mark.parametrize(
    "value,kind",
    [
        ("foo", _utils.ValueKind.VALUE),
        (1, _utils.ValueKind.VALUE),
        (1.0, _utils.ValueKind.VALUE),
        (b"123", _utils.ValueKind.VALUE),
        (Path("hello.txt"), _utils.ValueKind.VALUE),
        (True, _utils.ValueKind.VALUE),
        (False, _utils.ValueKind.VALUE),
        (Color.GREEN, _utils.ValueKind.VALUE),
        (Dataclass, _utils.ValueKind.VALUE),
        (Dataframe(), _utils.ValueKind.VALUE),
        (IntegerNode(123), _utils.ValueKind.VALUE),
        (DictConfig({}), _utils.ValueKind.VALUE),
        (ListConfig([]), _utils.ValueKind.VALUE),
        (AnyNode(123), _utils.ValueKind.VALUE),
        (UnionNode(123, Union[int, str]), _utils.ValueKind.VALUE),
        ("???", _utils.ValueKind.MANDATORY_MISSING),
        (IntegerNode("???"), _utils.ValueKind.MANDATORY_MISSING),
        (DictConfig("???"), _utils.ValueKind.MANDATORY_MISSING),
        (ListConfig("???"), _utils.ValueKind.MANDATORY_MISSING),
        (AnyNode("???"), _utils.ValueKind.MANDATORY_MISSING),
        (UnionNode("???", Union[int, str]), _utils.ValueKind.MANDATORY_MISSING),
        ("${foo.bar}", _utils.ValueKind.INTERPOLATION),
        ("ftp://${host}/path", _utils.ValueKind.INTERPOLATION),
        ("${func:foo}", _utils.ValueKind.INTERPOLATION),
        ("${func:a/b}", _utils.ValueKind.INTERPOLATION),
        ("${func:c:\\a\\b}", _utils.ValueKind.INTERPOLATION),
        ("${func:c:\\a\\b}", _utils.ValueKind.INTERPOLATION),
        param(
            IntegerNode("${func:c:\\a\\b}"),
            _utils.ValueKind.INTERPOLATION,
            id="integernode-interp",
        ),
        param(
            DictConfig("${func:c:\\a\\b}"),
            _utils.ValueKind.INTERPOLATION,
            id="dictconfig-interp",
        ),
        param(
            ListConfig("${func:c:\\a\\b}"),
            _utils.ValueKind.INTERPOLATION,
            id="listconfig-interp",
        ),
        param(
            AnyNode("${func:c:\\a\\b}"),
            _utils.ValueKind.INTERPOLATION,
            id="anynode-interp",
        ),
        param(
            UnionNode("${func:c:\\a\\b}", Union[int, str]),
            _utils.ValueKind.INTERPOLATION,
            id="unionnode-interp",
        ),
    ],
)
def test_value_kind(value: Any, kind: _utils.ValueKind) -> None:
    assert _utils.get_value_kind(value) == kind


def test_re_parent() -> None:
    def validate(cfg1: DictConfig) -> None:
        assert cfg1._get_parent() is None
        assert cfg1._get_node("str")._get_parent() == cfg1  # type: ignore
        assert cfg1._get_node("list")._get_parent() == cfg1  # type: ignore
        assert cfg1.list._get_node(0)._get_parent() == cfg1.list
        unode1 = cfg1._get_node("union")
        assert unode1._get_parent() == cfg1  # type: ignore
        assert unode1._value()._get_parent() == unode1  # type: ignore

    cfg = OmegaConf.create({})
    assert isinstance(cfg, DictConfig)
    cfg.str = StringNode("str")
    cfg.list = [1]
    cfg.union = UnionNode(123, Union[int, str])

    validate(cfg)

    cfg._get_node("str")._set_parent(None)  # type: ignore
    cfg._get_node("list")._set_parent(None)  # type: ignore
    cfg.list._get_node(0)._set_parent(None)  # type:ignore
    unode = cfg._get_node("union")
    unode._set_parent(None)  # type: ignore
    unode._value()._set_parent(None)  # type: ignore
    # noinspection PyProtectedMember
    cfg._re_parent()
    validate(cfg)


def test_get_class() -> None:
    name = "tests.examples.test_dataclass_example.SimpleTypes"
    assert _utils._get_class(name).__name__ == "SimpleTypes"
    with raises(ValueError):
        _utils._get_class("not_found")

    with raises(ModuleNotFoundError):
        _utils._get_class("foo.not_found")

    with raises(ImportError):
        _utils._get_class("tests.examples.test_dataclass_example.not_found")


@mark.parametrize(
    "key_type,expected_key_type",
    [
        (str, str),
        (Color, Color),
        (Any, Any),
    ],
)
@mark.parametrize(
    "value_type,expected_value_type",
    [
        (int, int),
        (str, str),
        (Color, Color),
        (Any, Any),
    ],
)
def test_get_key_value_types(
    key_type: Any, expected_key_type: Any, value_type: Any, expected_value_type: Any
) -> None:
    dt = Dict[key_type, value_type]  # type:ignore
    assert _utils.get_dict_key_value_types(dt) == (
        expected_key_type,
        expected_value_type,
    )


@mark.parametrize(
    "type_, is_primitive",
    [
        (int, True),
        (float, True),
        (bool, True),
        (bytes, True),
        (str, True),
        (Path, True),
        (NoneType, True),
        (Color, True),
        (list, False),
        (ListConfig, False),
        (dict, False),
        (DictConfig, False),
    ],
)
def test_is_primitive_type_annotation(type_: Any, is_primitive: bool) -> None:
    assert _utils.is_primitive_type_annotation(type_) == is_primitive


@mark.parametrize("optional", [False, True])
@mark.parametrize(
    "type_, include_module_name, expected",
    [
        (int, False, "int"),
        (int, True, "int"),
        (bool, False, "bool"),
        (bool, True, "bool"),
        (bytes, False, "bytes"),
        (bytes, True, "bytes"),
        (float, False, "float"),
        (float, True, "float"),
        (str, False, "str"),
        (str, True, "str"),
        (Path, False, "Path"),
        (Path, True, "pathlib.Path"),
        (Color, False, "Color"),
        (Color, True, "tests.Color"),
        (DictConfig, False, "DictConfig"),
        (DictConfig, True, "DictConfig"),
        (ListConfig, False, "ListConfig"),
        (ListConfig, True, "ListConfig"),
        (Dict[str, str], False, "Dict[str, str]"),
        (Dict[str, str], True, "Dict[str, str]"),
        (Dict[Color, int], False, "Dict[Color, int]"),
        (Dict[Color, int], True, "Dict[tests.Color, int]"),
        (Dict[str, Plugin], False, "Dict[str, Plugin]"),
        (Dict[str, Plugin], True, "Dict[str, tests.Plugin]"),
        (Dict[str, List[Plugin]], False, "Dict[str, List[Plugin]]"),
        (Dict[str, List[Plugin]], True, "Dict[str, List[tests.Plugin]]"),
        (dict, False, "dict"),
        (dict, True, "dict"),
        (List[str], False, "List[str]"),
        (List[str], True, "List[str]"),
        (List[Color], False, "List[Color]"),
        (List[Color], True, "List[tests.Color]"),
        (List[Dict[str, Color]], False, "List[Dict[str, Color]]"),
        (List[Dict[str, Color]], True, "List[Dict[str, tests.Color]]"),
        (list, False, "list"),
        (list, True, "list"),
        (Tuple[str], False, "Tuple[str]"),
        (Tuple[str], True, "Tuple[str]"),
        (Tuple[str, int], False, "Tuple[str, int]"),
        (Tuple[str, int], True, "Tuple[str, int]"),
        (Tuple[float, ...], False, "Tuple[float, ...]"),
        (Tuple[float, ...], True, "Tuple[float, ...]"),
        (tuple, False, "tuple"),
        (tuple, True, "tuple"),
        (Union[str, int, Color], False, "Union[str, int, Color]"),
        (Union[str, int, Color], True, "Union[str, int, tests.Color]"),
        (Union[int], False, "int"),
        (Union[int], True, "int"),
    ],
)
def test_type_str(
    type_: Any, include_module_name: bool, expected: str, optional: bool
) -> None:
    if optional:
        assert (
            _utils.type_str(Optional[type_], include_module_name=include_module_name)
            == f"Optional[{expected}]"
        )
    else:
        assert (
            _utils.type_str(type_, include_module_name=include_module_name) == expected
        )


def test_type_str_ellipsis() -> None:
    assert _utils.type_str(...) == "..."


@mark.parametrize(
    "type_, expected",
    [
        param(None, "NoneType", id="none"),
        param(NoneType, "NoneType", id="nonetype"),
        (Union[float, bool, None], "Optional[Union[float, bool]]"),
        (Union[float, bool, NoneType], "Optional[Union[float, bool]]"),
    ],
)
def test_type_str_nonetype(type_: Any, expected: str) -> None:
    assert _utils.type_str(type_) == expected


@mark.parametrize(
    "obj, expected",
    [
        param([], True, id="list"),
        param([1], True, id="list1"),
        param((), True, id="tuple"),
        param((1,), True, id="tuple1"),
        param({}, False, id="dict"),
        param(ListSubclass(), True, id="list_subclass"),
        param(Shape(10, 2, 3), True, id="namedtuple"),
    ],
)
def test_is_primitive_list(obj: Any, expected: bool) -> None:
    assert is_primitive_list(obj) == expected


@mark.parametrize(
    "obj, expected",
    [
        param({}, True, id="dict"),
        param({1: 2}, True, id="dict1"),
        param([], False, id="list"),
        param((), False, id="tuple"),
    ],
)
def test_is_primitive_dict(obj: Any, expected: bool) -> None:
    assert is_primitive_dict(obj) == expected


@mark.parametrize(
    "obj",
    [
        param(DictConfig({}), id="dictconfig"),
        param(ListConfig([]), id="listconfig"),
        param(DictSubclass(), id="dict_subclass"),
        param(Str2Int(), id="dict_subclass_dataclass"),
        param(User, id="user"),
        param(User("bond", 7), id="user"),
    ],
)
class TestIsPrimitiveContainerNegative:
    def test_is_primitive_list(self, obj: Any) -> None:
        assert not is_primitive_list(obj)

    def test_is_primitive_dict(self, obj: Any) -> None:
        assert not is_primitive_dict(obj)


@mark.parametrize(
    "type_, expected",
    [
        (Dict[str, int], True),
        (Dict[str, float], True),
        (Dict[bytes, bytes], True),
        (Dict[IllegalType, bool], True),
        (Dict[str, IllegalType], True),
        (Dict[int, Color], True),
        (Dict[Plugin, Plugin], True),
        (Dict[IllegalType, int], True),
        (Dict, True),
        (Str2Int, True),
        (Str2Int(), False),
        (User, False),
        (User(), False),
        (List, False),
        (dict, False),
        (DictConfig, False),
        (Any, False),
        (None, False),
        (NoneType, False),
    ],
)
def test_is_dict_annotation(type_: Any, expected: Any) -> Any:
    assert is_dict_annotation(type_=type_) == expected


@mark.parametrize(
    "type_, expected",
    [
        (List[int], True),
        (List[float], True),
        (List[bool], True),
        (List[bytes], True),
        (List[str], True),
        (List[Color], True),
        (List[Plugin], True),
        (List[IllegalType], True),
        (Dict, False),
        (List, True),
        (list, False),
        (tuple, False),
        (ListConfig, False),
    ],
)
def test_is_list_annotation(type_: Any, expected: Any) -> Any:
    assert is_list_annotation(type_=type_) == expected


@mark.parametrize(
    "type_, expected",
    [
        (Tuple[int], True),
        (Tuple[float], True),
        (Tuple[bool], True),
        (Tuple[str], True),
        (Tuple[Color], True),
        (Tuple[Plugin], True),
        (Tuple[IllegalType], True),
        (Dict, False),
        (List, False),
        (Tuple, True),
        (list, False),
        (dict, False),
        (tuple, False),
        (Any, False),
        (int, False),
        (User, False),
        (None, False),
        (NoneType, False),
    ],
)
def test_is_tuple_annotation(type_: Any, expected: Any) -> Any:
    assert is_tuple_annotation(type_=type_) == expected


@mark.parametrize(
    "input_, expected",
    [
        (Union[int, str], True),
        (Union[int, List[str]], True),
        (Optional[Union[int, str]], True),
        (Union[int, None], True),
        (Optional[int], True),
        (Any, False),
        (int, False),
        (User, False),
        (None, False),
        (NoneType, False),
    ],
)
def test_is_union_annotation(input_: Any, expected: bool) -> None:
    assert is_union_annotation(input_) == expected


@mark.skipif(sys.version_info < (3, 10), reason="requires Python 3.10 or newer")
def test_is_union_annotation_PEP604() -> None:
    if sys.version_info >= (3, 10):  # this if-statement is for mypy's benefit
        assert is_union_annotation(int | str)


@mark.parametrize(
    "input_, expected",
    [
        (Union[int, str], True),
        (Union[int, List[str]], False),
        (Union[int, Dict[str, int]], False),
        (Union[int, User], False),
        (Optional[Union[int, str]], True),
        (Union[int, None], True),
        (Optional[int], True),
        (Any, False),
        (int, False),
        (User, False),
        (None, False),
        (NoneType, False),
    ],
)
def test_is_supported_union_annotation(input_: Any, expected: bool) -> None:
    assert is_supported_union_annotation(input_) == expected


@mark.parametrize(
    "obj, expected",
    [
        # Unwrapped values
        param(10, Any, id="int"),
        param(10.0, Any, id="float"),
        param(True, Any, id="bool"),
        param(Color.RED, Any, id="enum"),
        param(b"binary", Any, id="bytes"),
        param(Path("hello.txt"), Any, id="path"),
        param("bar", Any, id="str"),
        param(None, Any, id="NoneType"),
        param({}, Any, id="dict"),
        param([], Any, id="List[Any]"),
        param(tuple(), Any, id="List[Any]"),
        param(ConcretePlugin(), Any, id="ConcretePlugin"),
        param(ConcretePlugin, Any, id="ConcretePlugin"),
        # Optional value nodes
        param(IntegerNode(10), Optional[int], id="IntegerNode"),
        param(FloatNode(10.0), Optional[float], id="FloatNode"),
        param(BooleanNode(True), Optional[bool], id="BooleanNode"),
        param(StringNode("bar"), Optional[str], id="StringNode"),
        param(BytesNode(b"binary"), Optional[bytes], id="BooleanNode"),
        param(
            EnumNode(enum_type=Color, value=Color.RED),
            Optional[Color],
            id="EnumNode[Color]",
        ),
        # Non-optional value nodes:
        param(IntegerNode(10, is_optional=False), int, id="IntegerNode"),
        param(FloatNode(10.0, is_optional=False), float, id="FloatNode"),
        param(BooleanNode(True, is_optional=False), bool, id="BooleanNode"),
        param(StringNode("bar", is_optional=False), str, id="StringNode"),
        param(BytesNode(b"binary", is_optional=False), bytes, id="BytesNode"),
        param(
            EnumNode(enum_type=Color, value=Color.RED, is_optional=False),
            Color,
            id="EnumNode[Color]",
        ),
        # DictConfig
        param(DictConfig(content={}), Any, id="DictConfig"),
        param(
            DictConfig(key_type=str, element_type=Color, content={}),
            Any,
            id="DictConfig[str,Color]",
        ),
        param(
            DictConfig(key_type=Color, element_type=int, content={}),
            Any,
            id="DictConfig[Color,int]",
        ),
        param(
            DictConfig(ref_type=Any, content=ConcretePlugin),
            Any,
            id="DictConfig[ConcretePlugin]_Any_reftype",
        ),
        param(
            DictConfig(content="???"),
            Any,
            id="DictConfig[Union[str, Enum], Any]_missing",
        ),
        param(
            DictConfig(content="???", element_type=int, key_type=str),
            Any,
            id="DictConfig[str, int]_missing",
        ),
        param(
            DictConfig(ref_type=Plugin, content=ConcretePlugin),
            Optional[Plugin],
            id="DictConfig[Plugin]",
        ),
        param(
            DictConfig(ref_type=Plugin, content=ConcretePlugin),
            Optional[Plugin],
            id="Plugin",
        ),
        # Non optional DictConfig
        param(
            DictConfig(ref_type=Plugin, content=ConcretePlugin, is_optional=False),
            Plugin,
            id="Plugin",
        ),
        # ListConfig
        param(ListConfig([]), Any, id="ListConfig[Any]"),
        param(ListConfig([], element_type=int), Any, id="ListConfig[int]"),
        param(ListConfig(content="???"), Any, id="ListConfig_missing"),
        param(
            ListConfig(content="???", element_type=int),
            Any,
            id="ListConfig[int]_missing",
        ),
        param(ListConfig(content=None), Any, id="ListConfig_none"),
        param(
            ListConfig(content=None, element_type=int),
            Any,
            id="ListConfig[int]_none",
        ),
    ],
)
def test_get_ref_type(obj: Any, expected: Any) -> None:
    assert _utils.get_type_hint(obj) == expected


@mark.parametrize(
    "obj, key, expected",
    [
        param({"foo": 10}, "foo", Any, id="dict"),
        param(User, "name", str, id="User.name"),
        param(User, "age", int, id="User.age"),
        param({"user": User}, "user", Any, id="user"),
        param(
            OmegaConf.structured(UnionAnnotations), "ubf", Union[bool, float], id="ubf"
        ),
        param(
            OmegaConf.structured(UnionAnnotations),
            "oubf",
            Optional[Union[bool, float]],
            id="oubf",
        ),
    ],
)
def test_get_node_ref_type(obj: Any, key: str, expected: Any) -> None:
    cfg = OmegaConf.create(obj)
    assert _utils.get_type_hint(cfg, key) == expected


def test_get_ref_type_error() -> None:
    with raises(ValueError):
        _utils.get_type_hint(AnyNode(), "foo")


@mark.parametrize(
    "value",
    [
        1,
        None,
        {"a": 0},
        [1, 2, 3],
    ],
)
def test_get_value_basic(value: Any) -> None:
    val_node = _node_wrap(
        value=value, ref_type=Any, parent=None, is_optional=True, key=None
    )
    result = _get_value(val_node)
    assert result == value


@mark.parametrize(
    "content",
    [{"a": 0, "b": 1}, "???", None, "${bar}"],
)
def test_get_value_container(content: Any) -> None:
    cfg = DictConfig({})
    cfg._set_value(content)
    assert _get_value(cfg) == content


@mark.parametrize(
    "node, expected",
    [
        param(AnyNode(123), 123, id="anynode"),
        param(IntegerNode(123), 123, id="integernode"),
        param(ListConfig([1, 2, 3]), ListConfig([1, 2, 3]), id="listconfig"),
        param(123, 123, id="int"),
        param("${a}", "${a}", id="raw-interp"),
        param(DictConfig("${a}"), "${a}", id="dict-interp"),
        param(AnyNode("${a}"), "${a}", id="any-interp"),
        param(IntegerNode("${a}"), "${a}", id="int-interp"),
        param(UnionNode("${a}", Union[int, str]), "${a}", id="union-interp"),
        param(DictConfig("???"), "???", id="dict-missing"),
        param(AnyNode("???"), "???", id="any-missing"),
        param(IntegerNode("???"), "???", id="int-missing"),
        param(UnionNode("???", Union[int, str]), "???", id="union-missing"),
        param(DictConfig(None), None, id="dict-none"),
        param(AnyNode(None), None, id="any-none"),
        param(IntegerNode(None), None, id="int-none"),
        param(UnionNode(None, Union[int, str]), None, id="union-none"),
        param(DictConfig({"foo": "bar"}), DictConfig({"foo": "bar"}), id="dictconfig"),
        param(UnionNode(123, Union[int, str]), 123, id="union[int]"),
    ],
)
def test_get_value_of_node_subclass(node: Node, expected: Any) -> None:
    result = _get_value(node)
    assert result == expected
    assert type(result) == type(expected)


def test_ensure_container_raises_ValueError() -> None:
    """Some values cannot be converted to a container.
    On these inputs, _ensure_container should raise a ValueError."""
    with raises(
        ValueError,
        match=re.escape(
            "Invalid input. Supports one of "
            + "[dict,list,DictConfig,ListConfig,dataclass,dataclass instance,attr class,attr class instance]"
        ),
    ):
        _ensure_container("abc")


def test_marker_string_representation() -> None:
    marker = Marker("marker")
    assert repr(marker) == "marker"


@mark.parametrize(
    ("key", "expected"),
    [
        ("", [""]),
        ("foo", ["foo"]),
        ("foo.bar", ["foo", "bar"]),
        ("foo[bar]", ["foo", "bar"]),
        (".foo", ["", "foo"]),
        ("..foo", ["", "", "foo"]),
        (".foo[bar]", ["", "foo", "bar"]),
        ("[foo]", ["foo"]),
        ("[foo][bar]", ["foo", "bar"]),
        (".[foo][bar]", ["", "foo", "bar"]),
        ("..[foo][bar]", ["", "", "foo", "bar"]),
        (
            "...a[b][c].d.e[f].g[h]",
            ["", "", "", "a", "b", "c", "d", "e", "f", "g", "h"],
        ),
    ],
)
def test_split_key(key: str, expected: List[str]) -> None:
    assert split_key(key) == expected


def test_nullcontext() -> None:
    with nullcontext() as x:
        assert x is None

    obj = object()
    with nullcontext(obj) as x:
        assert x is obj


@mark.parametrize("is_optional", [True, False])
@mark.parametrize(
    "fac",
    [
        (
            lambda is_optional, missing: StringNode(
                value="foo" if not missing else "???", is_optional=is_optional
            )
        ),
        (
            lambda is_optional, missing: IntegerNode(
                value=10 if not missing else "???", is_optional=is_optional
            )
        ),
        (
            lambda is_optional, missing: FloatNode(
                value=10.0 if not missing else "???", is_optional=is_optional
            )
        ),
        (
            lambda is_optional, missing: BooleanNode(
                value=True if not missing else "???", is_optional=is_optional
            )
        ),
        (
            lambda is_optional, missing: BytesNode(
                value=b"binary" if not missing else "???", is_optional=is_optional
            )
        ),
        (
            lambda is_optional, missing: EnumNode(
                enum_type=Color,
                value=Color.RED if not missing else "???",
                is_optional=is_optional,
            )
        ),
        (
            lambda is_optional, missing: ListConfig(
                content=[1, 2, 3] if not missing else "???", is_optional=is_optional
            )
        ),
        (
            lambda is_optional, missing: DictConfig(
                content={"foo": "bar"} if not missing else "???",
                is_optional=is_optional,
            )
        ),
        (
            lambda is_optional, missing: DictConfig(
                ref_type=ConcretePlugin,
                content=ConcretePlugin() if not missing else "???",
                is_optional=is_optional,
            )
        ),
        (
            lambda is_optional, missing: UnionNode(
                ref_type=Union[int, str],
                content=123 if not missing else "???",
                is_optional=is_optional,
            )
        ),
    ],
)
def test_is_optional(fac: Any, is_optional: bool) -> None:
    obj = fac(is_optional, False)
    assert _is_optional(obj) == is_optional

    cfg = OmegaConf.create({"node": obj})
    assert _is_optional(cfg, "node") == is_optional

    obj = fac(is_optional, True)
    assert _is_optional(obj) == is_optional

    cfg = OmegaConf.create({"node": obj})
    assert _is_optional(cfg, "node") == is_optional


@mark.parametrize(
    "type_",
    [
        param(lambda val=123: val, id="passthrough"),
        param(lambda val=123: AnyNode(val), id="any_node"),
        param(lambda val=123: IntegerNode(val), id="integer_node"),
        param(lambda val={}: DictConfig(val), id="dict_config"),
        param(lambda val=[]: ListConfig(val), id="list_config"),
        param(lambda val=123: UnionNode(val, Union[int, str]), id="union_node"),
    ],
)
class TestIndicators:
    @mark.parametrize(
        "input_, expected",
        [
            param("???", True, id="missing"),
            param("${interp}", False, id="interp"),
            param(None, False, id="none"),
            param("DEFAULT", False, id="default"),
        ],
    )
    def test_is_missing(
        self, type_: Callable[..., Any], input_: Any, expected: bool
    ) -> None:
        value = type_(input_) if input_ != "DEFAULT" else type_()
        assert _utils._is_missing_value(value) == expected

    @mark.parametrize(
        "input_, expected",
        [
            param("???", False, id="missing"),
            param("${interp}", True, id="interp"),
            param(None, False, id="none"),
            param("DEFAULT", False, id="default"),
        ],
    )
    def test_is_interpolation(
        self, type_: Callable[..., Any], input_: Any, expected: bool
    ) -> None:
        value = type_(input_) if input_ != "DEFAULT" else type_()
        assert _utils._is_interpolation(value) == expected

    @mark.parametrize(
        "input_, expected",
        [
            param("???", False, id="missing"),
            param("${interp}", False, id="interp"),
            param(None, True, id="none"),
            param("DEFAULT", False, id="default"),
        ],
    )
    def test_is_none(
        self, type_: Callable[..., Any], input_: Any, expected: bool
    ) -> None:
        value = type_(input_) if input_ != "DEFAULT" else type_()
        assert _utils._is_none(value) == expected

    @mark.parametrize(
        "input_, expected",
        [
            param("???", True, id="missing"),
            param("${interp}", True, id="interp"),
            param(None, True, id="none"),
            param("DEFAULT", False, id="default"),
        ],
    )
    def test_is_special(
        self, type_: Callable[..., Any], input_: Any, expected: bool
    ) -> None:
        value = type_(input_) if input_ != "DEFAULT" else type_()
        assert _utils._is_special(value) == expected


@mark.parametrize(
    "ref_type, expected_key_type, expected_element_type",
    [
        param(Dict, Any, Any, id="any"),
        param(Dict[Any, Any], Any, Any, id="any_explicit"),
        param(Dict[int, float], int, float, id="int_float"),
        param(Dict[Color, User], Color, User, id="color_user"),
        param(Dict[str, List[int]], str, List[int], id="list"),
        param(Dict[str, Dict[int, float]], str, Dict[int, float], id="dict"),
    ],
)
def test_get_dict_key_value_types(
    ref_type: Any, expected_key_type: Any, expected_element_type: Any
) -> None:
    key_type, element_type = get_dict_key_value_types(ref_type)
    assert key_type == expected_key_type
    assert element_type == expected_element_type


@mark.parametrize(
    "ref_type, expected_element_type",
    [
        param(List, Any, id="any"),
        param(List[Any], Any, id="any_explicit"),
        param(List[int], int, id="int"),
        param(List[User], User, id="user"),
        param(List[List[int]], List[int], id="list"),
        param(List[Dict[int, float]], Dict[int, float], id="dict"),
    ],
)
def test_get_list_element_type(ref_type: Any, expected_element_type: Any) -> None:
    assert get_list_element_type(ref_type) == expected_element_type


@mark.parametrize(
    "type_, expected_optional, expected_type",
    [
        param(int, False, int, id="int"),
        param(Any, True, Any, id="any"),
        param(Color, False, Color, id="color"),
        param(Optional[str], True, str, id="str"),
        param(Optional[Any], True, Any, id="o[any]"),
        param(Union[int, str], False, Union[int, str], id="int-str"),
        param(Union[str, int], False, Union[str, int], id="str-int"),
        param(Dict[str, int], False, Dict[str, int], id="dict[str,int]"),
        param(Dict, False, Dict, id="dict"),
        param(Dict[Any, Any], False, Dict[Any, Any], id="dict[any,any]"),
        param(Optional[Dict[str, int]], True, Dict[str, int], id="o[dict[str,int]]"),
        param(Optional[Dict], True, Dict, id="dict"),
        param(
            Dict[str, Optional[int]],
            False,
            Dict[str, Optional[int]],
            id="dict[str,o[int]]",
        ),
        param(Union[int, None], True, int, id="int-none"),
        param(Union[int, NoneType], True, int, id="int-nonetype"),
        param(Union[Optional[int], None], True, int, id="o[int]-none"),
        param(Union[Any, None], True, Any, id="any-none"),
        param(Union[None, int], True, int, id="none-int"),
        param(Union[None, None], True, NoneType, id="none-none"),
        param(Union[None, NoneType], True, NoneType, id="none-nonetype"),
        param(Union[None, Optional[None]], True, NoneType, id="none-o[none]"),
        param(None, True, NoneType, id="none"),
        param(NoneType, True, NoneType, id="nonetype"),
        param(Union[int, int], False, int, id="int-int"),
        param(Union[int, Optional[int]], True, int, id="int-o[int]"),
        param(Union[int], False, int, id="u[int]"),
        param(Union[Optional[int]], True, int, id="u[o[int]]"),
        param(Optional[Union[int]], True, int, id="o[u[int]]"),
        param(Union[int, Optional[str]], True, Union[int, str], id="int-o[str]"),
        param(Optional[Optional[int]], True, int, id="o[o[int]]"),
        param(Optional[Optional[Any]], True, Any, id="o[o[any]]"),
        param(User, False, User, id="user"),
        param(Optional[User], True, User, id="o[user]"),
        param(Union[User, int], False, Union[User, int], id="user-int"),
        param(Optional[Union[User, int]], True, Union[User, int], id="o[user-int]"),
        param(Union[Optional[User], int], True, Union[User, int], id="o[user]-int"),
        param(Union[User, Optional[int]], True, Union[User, int], id="user-o[int]"),
        param(Optional[Union[int, str]], True, Union[int, str], id="o[u[int-str]]"),
        param(Union[Optional[int], str], True, Union[int, str], id="u[o[int]-str]]"),
        param(
            Optional[Union[Optional[int], str]],
            True,
            Union[int, str],
            id="o[u[o[int]-str]]]",
        ),
        param(
            Union[Optional[int], Optional[str]],
            True,
            Union[int, str],
            id="u[o[int]-o[str]]]",
        ),
        param(Union[int, str, None], True, Union[int, str], id="u[int-str-none]"),
        param(Union[int, str, None], True, Union[int, str], id="u[int-str-nonetype]"),
        param(
            Union[User, Union[int, str]],
            False,
            Union[User, int, str],
            id="user-[int-str]",
        ),
        param(
            Union[User, NoneType, Union[int, str]],
            True,
            Union[User, int, str],
            id="user-nonetype-[int-str]",
        ),
        param(
            Union[User, None, Union[int, str]],
            True,
            Union[User, int, str],
            id="user-none-[int-str]",
        ),
        param(
            Union[User, None, Union[Optional[int], str]],
            True,
            Union[User, int, str],
            id="user-none-[o[int]-str]",
        ),
        param(
            Union[User, Union[Optional[int], str]],
            True,
            Union[User, int, str],
            id="user-none-[o[int]-str]",
        ),
        param(
            Union[float, bool, None], True, Union[float, bool], id="u[float-bool-none]"
        ),
        param(
            Union[float, bool, NoneType],
            True,
            Union[float, bool],
            id="u[float-bool-nonetype]",
        ),
    ],
)
def test_resolve_optional(
    type_: Any, expected_optional: bool, expected_type: Any
) -> None:
    resolved_optional, resolved_type = _resolve_optional(type_)
    assert resolved_optional == expected_optional
    assert resolved_type == expected_type


@mark.skipif(sys.version_info < (3, 10), reason="requires Python 3.10 or newer")
def test_resolve_optional_support_pep_604() -> None:
    if sys.version_info >= (3, 10):  # this if-statement is for mypy's benefit
        assert _resolve_optional(int | str) == (False, Union[int, str])
        assert _resolve_optional(Optional[int | str]) == (True, Union[int, str])
        assert _resolve_optional(int | Optional[str]) == (True, Union[int, str])
        assert _resolve_optional(int | Union[str, float]) == (
            False,
            Union[int, str, float],
        )
        assert _resolve_optional(int | Union[str, Optional[float]]) == (
            True,
            Union[int, str, float],
        )
        assert _resolve_optional(int | str | None) == (True, Union[int, str])
        assert _resolve_optional(int | str | NoneType) == (True, Union[int, str])



import copy
import functools
import re
import sys
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, Tuple, Type, Union

from pytest import fixture, mark, param, raises

from omegaconf import (
    AnyNode,
    BooleanNode,
    BytesNode,
    DictConfig,
    EnumNode,
    FloatNode,
    IntegerNode,
    ListConfig,
    Node,
    OmegaConf,
    PathNode,
    StringNode,
    UnionNode,
    ValueNode,
)
from omegaconf._utils import BUILTIN_VALUE_TYPES, type_str
from omegaconf.errors import (
    InterpolationToMissingValueError,
    UnsupportedValueType,
    ValidationError,
)
from omegaconf.nodes import InterpolationResultNode
from tests import Color, Enum1, IllegalType, User


def _build_union(content: Any) -> UnionNode:
    return UnionNode(content=content, ref_type=Union[(Enum, *BUILTIN_VALUE_TYPES)])


# testing valid conversions
@mark.parametrize(
    "type_,input_,output_",
    [
        # string
        (StringNode, "abc", "abc"),
        (StringNode, 100, "100"),
        (StringNode, Color.RED, "Color.RED"),
        # integer
        (IntegerNode, 10, 10),
        (IntegerNode, "10", 10),
        (IntegerNode, -100, -100),
        (IntegerNode, "-100", -100),
        # float
        (FloatNode, float("inf"), float("inf")),
        # Yes, we treat nan as equal to nan in OmegaConf
        (FloatNode, float("nan"), float("nan")),
        (FloatNode, 10, 10.0),
        (FloatNode, 10.1, 10.1),
        (FloatNode, "10.2", 10.2),
        (FloatNode, "10e-3", 10e-3),
        # bytes
        (BytesNode, b"binary", b"binary"),
        (BytesNode, b"\xf0\xf1\xf2", b"\xf0\xf1\xf2"),
        # bool true
        (BooleanNode, True, True),
        (BooleanNode, "Y", True),
        (BooleanNode, "true", True),
        (BooleanNode, "Yes", True),
        (BooleanNode, "On", True),
        (BooleanNode, "1", True),
        (BooleanNode, 100, True),
        # bool false
        (BooleanNode, False, False),
        (BooleanNode, "N", False),
        (BooleanNode, "false", False),
        (BooleanNode, "No", False),
        (BooleanNode, "Off", False),
        (BooleanNode, None, None),
        (BooleanNode, "0", False),
        (BooleanNode, 0, False),
        # any
        (AnyNode, 3, 3),
        (AnyNode, 3.14, 3.14),
        (AnyNode, False, False),
        (AnyNode, b"\xf0\xf1\xf2", b"\xf0\xf1\xf2"),
        (AnyNode, Color.RED, Color.RED),
        (AnyNode, None, None),
        # Enum node
        (lambda v: EnumNode(enum_type=Color, value=v), Color.RED, Color.RED),
        (lambda v: EnumNode(enum_type=Color, value=v), "Color.RED", Color.RED),
        (lambda v: EnumNode(enum_type=Color, value=v), "RED", Color.RED),
        (lambda v: EnumNode(enum_type=Color, value=v), 1, Color.RED),
        # Path node
        (PathNode, "hello.txt", Path("hello.txt")),
        (PathNode, Path("hello.txt"), Path("hello.txt")),
        # Union node
        param(_build_union, "abc", "abc", id="union-str"),
        param(_build_union, 10, 10, id="union-int"),
        param(_build_union, 10.1, 10.1, id="union-float"),
        param(_build_union, float("inf"), float("inf"), id="union-inf"),
        param(_build_union, b"binary\xf0\xf1", b"binary\xf0\xf1", id="union-bytes"),
        param(
            _build_union,
            True,
            True,
            marks=mark.skipif(
                sys.version_info < (3, 7),
                reason="python3.6 treats Union[int, bool] as equivalent to Union[int]",
            ),
            id="union-bool",
        ),
        param(_build_union, None, None, id="union-none"),
        param(_build_union, Color.RED, Color.RED, id="union-enum"),
    ],
)
def test_valid_inputs(type_: type, input_: Any, output_: Any) -> None:
    node = type_(input_)
    assert node == output_
    assert node == node
    assert not (node != output_)
    assert not (node != node)
    assert str(node) == str(output_)
    assert repr(node) == repr(output_)


# testing invalid conversions
@mark.parametrize(
    "type_,input_",
    [
        (IntegerNode, "abc"),
        (IntegerNode, "-abc"),
        (IntegerNode, 10.1),
        (IntegerNode, "-1132c"),
        (IntegerNode, Color.RED),
        (IntegerNode, b"123"),
        (FloatNode, "abc"),
        (FloatNode, Color.RED),
        (FloatNode, b"10.1"),
        (BytesNode, "abc"),
        (BytesNode, 23),
        (BytesNode, Color.RED),
        (BytesNode, 3.14),
        (BytesNode, True),
        (BooleanNode, "Nope"),
        (BooleanNode, "Yup"),
        (BooleanNode, Color.RED),
        (BooleanNode, b"True"),
        (IntegerNode, [1, 2]),
        (IntegerNode, ListConfig([1, 2])),
        (IntegerNode, {"foo": "var"}),
        (IntegerNode, b"10"),
        (IntegerNode, DictConfig({"foo": "var"})),
        (BytesNode, [1, 2]),
        (BytesNode, ListConfig([1, 2])),
        (BytesNode, {"foo": "var"}),
        (BytesNode, DictConfig({"foo": "var"})),
        (BooleanNode, [1, 2]),
        (BooleanNode, ListConfig([1, 2])),
        (BooleanNode, {"foo": "var"}),
        (BooleanNode, DictConfig({"foo": "var"})),
        (FloatNode, [1, 2]),
        (FloatNode, ListConfig([1, 2])),
        (FloatNode, {"foo": "var"}),
        (FloatNode, DictConfig({"foo": "var"})),
        (StringNode, [1, 2]),
        (StringNode, ListConfig([1, 2])),
        (StringNode, {"foo": "var"}),
        (StringNode, b"\xf0\xf1\xf2"),
        (FloatNode, DictConfig({"foo": "var"})),
        (AnyNode, [1, 2]),
        (AnyNode, ListConfig([1, 2])),
        (AnyNode, {"foo": "var"}),
        (AnyNode, DictConfig({"foo": "var"})),
        (AnyNode, IllegalType()),
        (partial(EnumNode, Color), "Color.TYPO"),
        (partial(EnumNode, Color), "TYPO"),
        (partial(EnumNode, Color), Enum1.FOO),
        (partial(EnumNode, Color), "Enum1.RED"),
        (partial(EnumNode, Color), 1000000),
        (partial(EnumNode, Color), 1.0),
        (partial(EnumNode, Color), b"binary"),
        (partial(EnumNode, Color), True),
        (partial(EnumNode, Color), [1, 2]),
        (partial(EnumNode, Color), {"foo": "bar"}),
        (partial(EnumNode, Color), ListConfig([1, 2])),
        (partial(EnumNode, Color), DictConfig({"foo": "bar"})),
        (PathNode, 1.0),
        (PathNode, ["hello.txt"]),
    ],
)
def test_invalid_inputs(type_: type, input_: Any) -> None:
    empty_node = type_()

    with raises(ValidationError):
        empty_node._set_value(input_)
    with raises(ValidationError):
        type_(input_)


@mark.parametrize(
    "optional", [param(True, id="optional"), param(False, id="not_optional")]
)
@mark.parametrize(
    "input_",
    [
        param("???", id="missing"),
        param("${interp}", id="interp"),
        param(None, id="none"),
    ],
)
@mark.parametrize(
    "flags", [param(None, id="convert"), param({"convert": False}, id="no_convert")]
)
class TestValueNodeSpecial:
    @mark.parametrize(
        "type_",
        [
            param(IntegerNode, id="int"),
            param(FloatNode, id="float"),
            param(BytesNode, id="bytes"),
            param(BooleanNode, id="bool"),
            param(StringNode, id="str"),
            param(partial(EnumNode, Color), id="enum"),
            param(PathNode, id="path"),
        ],
    )
    def test_creation_special(
        self, type_: Callable[..., Node], input_: Any, optional: bool, flags: Any
    ) -> None:
        if input_ is None and not optional:
            with raises(ValidationError):
                type_(input_, is_optional=optional, flags=flags)
        else:
            node = type_(input_, is_optional=optional, flags=flags)
            assert node._value() == input_

    @mark.parametrize(
        "builds_node",
        [
            param(partial(IntegerNode, 123), id="int"),
            param(partial(FloatNode, 10.1), id="float"),
            param(partial(BytesNode, b"binary"), id="bytes"),
            param(partial(BooleanNode, True), id="bool"),
            param(partial(StringNode, "abc"), id="str"),
            param(partial(EnumNode, Color, value=Color.RED), id="enum"),
            param(partial(PathNode, Path("hello.txt")), id="path"),
        ],
    )
    def test_set_value_special(
        self, builds_node: Callable[..., Node], input_: Any, optional: bool, flags: Any
    ) -> None:
        node = builds_node(is_optional=optional, flags=flags)
        if input_ is None and not optional:
            with raises(ValidationError):
                node._set_value(input_)
        else:
            node._set_value(input_)
            assert node._value() == input_


@mark.parametrize(
    "input_",
    [
        param(123, id="123"),
        param(10.1, id="10.1"),
        param(b"binary", id="binary"),
        param(True, id="true"),
        param("abc", id="abc"),
        param("RED", id="red_str"),
        param("123", id="123_str"),
        param("10.1", id="10.1_str"),
        param(Color.RED, id="Color.RED"),
        param(Enum1.FOO, id="Enum1.FOO"),
        param(Path("hello.txt"), id="path"),
        param(object(), id="object"),
    ],
)
@mark.parametrize(
    "type_, legal_type",
    [
        param(IntegerNode, int, id="integer_node"),
        param(FloatNode, float, id="float_node"),
        param(BytesNode, bytes, id="bytes_node"),
        param(BooleanNode, bool, id="boolean_node"),
        param(StringNode, str, id="string_node"),
        param(partial(EnumNode, Color), Color, id="enum_node"),
        param(partial(EnumNode, Enum), lambda t: issubclass(t, Enum), id="enum_node"),
        param(PathNode, lambda t: issubclass(t, Path), id="path_node"),
    ],
)
class TestNodeConvertFalse:
    @fixture
    def input_is_compatible_with_node(
        self, input_: Any, legal_type: Union[Type[Any], Callable[[Type[Any]], bool]]
    ) -> bool:
        input_type = type(input_)
        return (
            input_type is legal_type
            if isinstance(legal_type, type)
            else legal_type(input_type)
        )

    @mark.parametrize(
        "optional", [param(True, id="optional"), param(False, id="not_optional")]
    )
    def test_instantiate_with_value(
        self,
        type_: Callable[..., ValueNode],
        input_: Any,
        input_is_compatible_with_node: bool,
        optional: bool,
    ) -> None:
        if input_is_compatible_with_node:
            node = type_(input_, is_optional=optional, flags={"convert": False})
            assert node._value() == input_
        else:
            with raises(ValidationError):
                type_(input_, is_optional=optional, flags={"convert": False})

    def test_set_value(
        self,
        type_: Callable[..., ValueNode],
        input_: Any,
        input_is_compatible_with_node: bool,
    ) -> None:
        node = type_(flags={"convert": False})
        if input_is_compatible_with_node:
            node._set_value(input_, flags={"convert": False})
            assert node._value() == input_
        else:
            with raises(ValidationError):
                node._set_value(input_, flags={"convert": False})


@mark.parametrize(
    "input_, expected_type",
    [
        ({}, DictConfig),
        ([], ListConfig),
        (5, AnyNode),
        (5.0, AnyNode),
        (True, AnyNode),
        (False, AnyNode),
        ("str", AnyNode),
        (b"\xf0\xf1\xf2", AnyNode),
        (Path("hello.txt"), AnyNode),
    ],
)
def test_assigned_value_node_type(input_: type, expected_type: Any) -> None:
    c = OmegaConf.create()
    assert isinstance(c, DictConfig)
    c.foo = input_
    assert type(c._get_node("foo")) == expected_type


# dict
def test_dict_any() -> None:
    c = OmegaConf.create()
    assert isinstance(c, DictConfig)
    # default type is Any
    c.foo = 10
    c[Enum1.FOO] = "bar"

    assert c.foo == 10
    assert type(c._get_node("foo")) == AnyNode
    c.foo = "string"
    assert c.foo == "string"

    assert type(c._get_node(Enum1.FOO)) == AnyNode


def test_dict_integer_1() -> None:
    c = OmegaConf.create()
    assert isinstance(c, DictConfig)
    c.foo = IntegerNode(10)
    assert type(c._get_node("foo")) == IntegerNode
    assert c.foo == 10


# list
def test_list_any() -> None:
    c = OmegaConf.create([])
    assert isinstance(c, ListConfig)
    # default type is Any
    c.append(10)
    assert c[0] == 10
    assert type(c._get_node(0)) == AnyNode
    c[0] = "string"
    assert c[0] == "string"


def test_list_integer() -> None:
    val = 10
    c = OmegaConf.create([])
    assert isinstance(c, ListConfig)
    c.append(IntegerNode(val))
    assert type(c._get_node(0)) == IntegerNode
    assert c.get(0) == val


def test_list_integer_rejects_string() -> None:
    c = OmegaConf.create([])
    assert isinstance(c, ListConfig)
    c.append(IntegerNode(10))
    assert c.get(0) == 10
    with raises(ValidationError):
        c[0] = "string"
    assert c[0] == 10
    assert type(c._get_node(0)) == IntegerNode


# Test merge raises validation error
@mark.parametrize(
    "c1, c2",
    [
        (dict(a=IntegerNode(10)), dict(a="str")),
        (dict(a=IntegerNode(10)), dict(a=StringNode("str"))),
        (dict(a=10, b=IntegerNode(10)), dict(a=20, b="str")),
        (dict(foo=dict(bar=IntegerNode(10))), dict(foo=dict(bar="str"))),
    ],
)
def test_merge_validation_error(c1: Dict[str, Any], c2: Dict[str, Any]) -> None:
    conf1 = OmegaConf.create(c1)
    conf2 = OmegaConf.create(c2)
    with raises(ValidationError):
        OmegaConf.merge(conf1, conf2)
    # make sure that conf1 and conf2 were not modified
    assert conf1 == OmegaConf.create(c1)
    assert conf2 == OmegaConf.create(c2)


@mark.parametrize(
    "type_,valid_value, invalid_value",
    [
        (IntegerNode, 1, "invalid"),
        (FloatNode, 3.1415, "invalid"),
        (BooleanNode, True, "invalid"),
        (AnyNode, "aaa", None),
        (StringNode, "blah", None),
        (BytesNode, b"foobar", None),
        (PathNode, Path("hello.txt"), None),
    ],
)
def test_accepts_mandatory_missing(
    type_: type, valid_value: Any, invalid_value: Any
) -> None:
    node = type_()
    node._set_value("???")
    assert node._value() == "???"

    conf = OmegaConf.create({"foo": node})
    assert isinstance(conf, DictConfig)
    assert "foo" not in conf
    assert type(conf._get_node("foo")) == type_

    conf.foo = valid_value
    # make sure valid assignment does not change the type
    assert type(conf._get_node("foo")) == type_
    assert "foo" in conf
    assert conf.foo == valid_value

    if invalid_value is not None:
        with raises(ValidationError):
            conf.foo = invalid_value


@mark.parametrize(
    "type_",
    [
        BooleanNode,
        BytesNode,
        PathNode,
        EnumNode,
        FloatNode,
        IntegerNode,
        StringNode,
        AnyNode,
    ],
)
@mark.parametrize(
    "values, success_map",
    [
        param(
            # Integers
            (1, 10, -10),
            {
                "BooleanNode": True,
                "IntegerNode": int,
                "FloatNode": float,
                "StringNode": str,
                "AnyNode": copy.copy,
            },
            id="integers",
        ),
        param(
            # Integer strings
            ("1",),
            {
                "BooleanNode": True,
                "IntegerNode": int,
                "FloatNode": float,
                "StringNode": str,
                "AnyNode": copy.copy,
                "PathNode": Path,
            },
            id="integer-strings",
        ),
        param(
            # Floats
            (1.0, float("inf"), float("-inf"), 10e-3),
            {"FloatNode": float, "StringNode": str, "AnyNode": copy.copy},
            id="floats",
        ),
        param(
            # Floaty strings
            ("1.0", "10e-3"),
            {
                "FloatNode": float,
                "StringNode": str,
                "AnyNode": copy.copy,
                "PathNode": Path,
            },
            id="floaty-strings",
        ),
        param(
            # Booleans
            (True, False),
            {
                "BooleanNode": copy.copy,
                "StringNode": str,
                "AnyNode": copy.copy,
            },
            id="booleans",
        ),
        param(
            # Trueish strings
            ("Y", "true", "yes", "on"),
            {
                "BooleanNode": True,
                "StringNode": str,
                "AnyNode": copy.copy,
                "PathNode": Path,
            },
            id="trueish-strings",
        ),
        param(
            # Falsey strings
            ("N", "false", "no", "off"),
            {
                "BooleanNode": False,
                "StringNode": str,
                "AnyNode": copy.copy,
                "PathNode": Path,
            },
            id="falsey-strings",
        ),
        param(
            # Falsy integer
            (0,),
            {
                "BooleanNode": False,
                "IntegerNode": 0,
                "FloatNode": 0.0,
                "StringNode": str,
                "AnyNode": copy.copy,
            },
            id="falsey-integer",
        ),
        param(
            # Falsy integer string
            ("0",),
            {
                "BooleanNode": False,
                "IntegerNode": 0,
                "FloatNode": 0.0,
                "StringNode": str,
                "AnyNode": copy.copy,
                "PathNode": Path,
            },
            id="falsey-integer-string",
        ),
        param(
            # Binary data
            (b"binary",),
            {
                "BytesNode": b"binary",
                "AnyNode": copy.copy,
            },
            id="binary-data",
        ),
        param(
            # pathlib.Path data
            (Path("hello.txt"),),
            {
                "PathNode": copy.copy,
                "AnyNode": copy.copy,
                "StringNode": str,
            },
            id="path-data",
        ),
    ],
)
def test_legal_assignment(
    type_: type, values: Any, success_map: Dict[Any, Dict[str, Any]]
) -> None:

    if not isinstance(values, (list, tuple)):
        values = [values]

    for value in values:
        if type_.__name__ in success_map.keys():
            expected = success_map[type_.__name__]
            if callable(expected):
                expected = expected(value)
            node = type_(value)
            assert node._value() == expected
        else:
            with raises(ValidationError):
                type_(value)


@mark.parametrize(
    "node,value",
    [
        (IntegerNode(), "foo"),
        (BooleanNode(), "foo"),
        (FloatNode(), "foo"),
        (BytesNode(), "foo"),
        (PathNode(), 123),
        (EnumNode(enum_type=Enum1), "foo"),
    ],
)
def test_illegal_assignment(node: ValueNode, value: Any) -> None:
    with raises(ValidationError):
        node._set_value(value)


@mark.parametrize(
    "node_type",
    [
        BooleanNode,
        BytesNode,
        PathNode,
        EnumNode,
        FloatNode,
        IntegerNode,
        StringNode,
        AnyNode,
    ],
)
@mark.parametrize(
    "enum_type, values, success_map",
    [
        (
            Enum1,
            (Enum1.FOO, "Enum1.FOO", "FOO", 1),
            {EnumNode: Enum1.FOO, AnyNode: copy.copy, StringNode: str},
        )
    ],
)
def test_legal_assignment_enum(
    node_type: Type[EnumNode],
    enum_type: Type[Enum],
    values: Tuple[Any, ...],
    success_map: Dict[Any, Any],
) -> None:
    assert isinstance(values, (list, tuple))

    for value in values:
        if node_type in success_map.keys():
            expected = success_map[node_type]
            if callable(expected):
                expected = expected(value)
            node = node_type(enum_type)
            node._set_value(value)
            assert node._value() == expected
        else:
            with raises(ValidationError):
                node_type(enum_type)


@mark.parametrize(
    "obj",
    [
        StringNode(),
        StringNode(value="foo"),
        StringNode(value="foo", is_optional=False),
        BytesNode(value=b"\xf0\xf1\xf2"),
        PathNode(value=Path("hello.txt")),
        BooleanNode(value=True),
        IntegerNode(value=10),
        FloatNode(value=10.0),
        UnionNode(10.0, Union[float, bool]),
        OmegaConf.create({}),
        OmegaConf.create([]),
        OmegaConf.create({"foo": "foo"}),
    ],
)
def test_deepcopy(obj: Any) -> None:
    cp = copy.deepcopy(obj)
    assert cp == obj
    assert id(cp) != id(obj)
    assert obj.__dict__.keys() == cp.__dict__.keys()
    for k in obj.__dict__.keys():
        assert obj.__dict__[k] == cp.__dict__[k]


@mark.parametrize(
    "node, value, expected",
    [
        (StringNode(), None, True),
        (StringNode(), 100, False),
        (StringNode("foo"), "foo", True),
        (StringNode("100"), 100, False),
        (StringNode("???"), "???", True),
        (StringNode(None), None, True),
        (StringNode("abc"), None, False),
        (StringNode("${interp}"), "${interp}", True),
        (StringNode("${interp}"), "${different_interp}", False),
        (IntegerNode(), 1, False),
        (IntegerNode(1), 1, True),
        (IntegerNode(1), "1", False),
        (IntegerNode(1), "foo", False),
        (IntegerNode(1), "foo", False),
        (IntegerNode("???"), "???", True),
        (IntegerNode(None), None, True),
        (IntegerNode("${interp}"), "${interp}", True),
        (IntegerNode("${interp}"), "${different_interp}", False),
        (FloatNode(), 1, False),
        (FloatNode(), None, True),
        (FloatNode(1.0), None, False),
        (FloatNode(1.0), 1.0, True),
        (FloatNode(1), 1, True),
        (FloatNode(1), "1", False),
        (FloatNode(1.0), "1.0", False),
        (FloatNode(1.0), "foo", False),
        (FloatNode("???"), "???", True),
        (BytesNode(), None, True),
        (BytesNode(), b"binary", False),
        (BytesNode(b"binary"), b"binary", True),
        (PathNode(), None, True),
        (PathNode(), Path("hello.txt"), False),
        (PathNode(Path("hello.txt")), Path("hello.txt"), True),
        (PathNode(Path("hello.txt")), "hello.txt", False),
        (BooleanNode(), True, False),
        (BooleanNode(), False, False),
        (BooleanNode(), None, True),
        (BooleanNode(True), None, False),
        (BooleanNode(True), False, False),
        (BooleanNode(False), False, True),
        (AnyNode(value=1), AnyNode(value=1), True),
        (AnyNode(value=1), 1, True),
        (AnyNode(value=1), AnyNode(value="1"), False),
        (AnyNode(value=1), "1", False),
        (EnumNode(enum_type=Enum1), Enum1.BAR, False),
        (EnumNode(enum_type=Enum1), EnumNode(Enum1), True),
        (EnumNode(enum_type=Enum1), "nope", False),
        (
            EnumNode(enum_type=Enum1, value=Enum1.BAR),
            EnumNode(enum_type=Enum1, value=Enum1.BAR),
            True,
        ),
        (
            EnumNode(enum_type=Enum1, value=Enum1.BAR),
            EnumNode(enum_type=Enum1, value=Enum1.FOO),
            False,
        ),
        (EnumNode(enum_type=Enum1, value=Enum1.BAR), Enum1.BAR, True),
        (EnumNode(enum_type=Enum1, value=Enum1.BAR), Enum1.FOO, False),
        (EnumNode(enum_type=Enum1, value=Enum1.BAR), "Enum1.BAR", False),
        (EnumNode(enum_type=Enum1, value="???"), "???", True),
        (EnumNode(enum_type=Enum1, value=None), Enum1.BAR, False),
        (EnumNode(enum_type=Enum1, value="${interp}"), "${interp}", True),
        (EnumNode(enum_type=Enum1, value="${interp}"), "${different_interp}", False),
        (InterpolationResultNode("foo"), "foo", True),
        (InterpolationResultNode("${foo}"), "${foo}", True),
        (InterpolationResultNode("${foo"), "${foo", True),
        (InterpolationResultNode(None), None, True),
        (InterpolationResultNode(1), 1, True),
        (InterpolationResultNode(1), "1", False),
        (InterpolationResultNode(1.0), 1.0, True),
        (InterpolationResultNode(True), True, True),
        (InterpolationResultNode(Color.RED), Color.RED, True),
        (InterpolationResultNode({"a": 0, "b": 1}), {"a": 0, "b": 1}, True),
        (InterpolationResultNode([0, None, True]), [0, None, True], True),
        (InterpolationResultNode("foo"), 100, False),
        (InterpolationResultNode(100), "foo", False),
        (InterpolationResultNode("???"), "???", True),
        (InterpolationResultNode(None), None, True),
        (InterpolationResultNode(None), 100, False),
        (UnionNode(100, Union[int, bytes]), UnionNode(100, Union[int, bytes]), True),
        (UnionNode(100, Union[int, bytes]), 100, True),
        (UnionNode(100, Union[int, bytes]), IntegerNode(100), True),
        (UnionNode(100, Union[int, bytes]), AnyNode(100), True),
        (
            UnionNode("???", Union[int, bytes]),
            UnionNode("???", Union[int, bytes]),
            True,
        ),
        (UnionNode("???", Union[int, bytes]), "???", True),
        (UnionNode("???", Union[int, bytes]), IntegerNode("???"), True),
        (UnionNode("???", Union[int, bytes]), AnyNode("???"), True),
        (UnionNode(None, Union[int, bytes]), UnionNode(None, Union[int, bytes]), True),
        (UnionNode(None, Union[int, bytes]), None, True),
        (UnionNode(None, Union[int, bytes]), IntegerNode(None), True),
        (UnionNode(None, Union[int, bytes]), AnyNode(None), True),
        (
            UnionNode("${interp}", Union[int, bytes]),
            UnionNode("${interp}", Union[int, bytes]),
            True,
        ),
        (UnionNode("${interp}", Union[int, bytes]), "${interp}", True),
        (UnionNode("${interp}", Union[int, bytes]), IntegerNode("${interp}"), True),
        (UnionNode("${interp}", Union[int, bytes]), AnyNode("${interp}"), True),
        (UnionNode(100, Union[int, bytes]), UnionNode(999, Union[int, bytes]), False),
        (UnionNode(100, Union[int, bytes]), 999, False),
        (UnionNode(100, Union[int, bytes]), IntegerNode(999), False),
        (UnionNode(100, Union[int, bytes]), AnyNode(999), False),
        (UnionNode("???", Union[int, bytes]), UnionNode(999, Union[int, bytes]), False),
        (UnionNode("???", Union[int, bytes]), 999, False),
        (UnionNode("???", Union[int, bytes]), IntegerNode(999), False),
        (UnionNode("???", Union[int, bytes]), AnyNode(999), False),
        (UnionNode(None, Union[int, bytes]), UnionNode(999, Union[int, bytes]), False),
        (UnionNode(None, Union[int, bytes]), 999, False),
        (UnionNode(None, Union[int, bytes]), IntegerNode(999), False),
        (UnionNode(None, Union[int, bytes]), AnyNode(999), False),
        (
            UnionNode("${interp}", Union[int, bytes]),
            UnionNode(999, Union[int, bytes]),
            False,
        ),
        (UnionNode("${interp}", Union[int, bytes]), 999, False),
        (UnionNode("${interp}", Union[int, bytes]), IntegerNode(999), False),
        (UnionNode("${interp}", Union[int, bytes]), AnyNode(999), False),
    ],
)
def test_eq(node: ValueNode, value: Any, expected: Any) -> None:
    assert (node == value) == expected
    assert (node != value) != expected
    assert (value == node) == expected
    assert (value != node) != expected

    # Check hash except for unhashable types (dict/list).
    if not isinstance(value, (dict, list)):
        assert (node.__hash__() == value.__hash__()) == expected


@mark.parametrize(
    "value", ["a_str", 1, 3.14, True, None, Enum1.FOO, b"binary", Path("hello.txt")]
)
def test_set_anynode_with_primitive_type(value: Any) -> None:
    cfg = OmegaConf.create({"a": 5})
    a_before = cfg._get_node("a")
    cfg.a = value
    # changing anynode's value with a primitive type should set value
    assert id(cfg._get_node("a")) == id(a_before)
    assert cfg.a == value


@mark.parametrize(
    "value, container_type",
    [
        (ListConfig(content=[1, 2]), ListConfig),
        ([1, 2], ListConfig),
        (DictConfig(content={"foo": "var"}), DictConfig),
        ({"foo": "var"}, DictConfig),
    ],
)
def test_set_anynode_with_container(value: Any, container_type: Any) -> None:
    cfg = OmegaConf.create({"a": 5})
    a_before = cfg._get_node("a")
    cfg.a = value
    # changing anynode's value with a container should wrap a new node
    assert id(cfg._get_node("a")) != id(a_before)
    assert isinstance(cfg.a, container_type)
    assert cfg.a == value


def test_set_anynode_with_illegal_type() -> None:
    cfg = OmegaConf.create({"a": 5})
    with raises(ValidationError):
        cfg.a = IllegalType()


def test_set_valuenode() -> None:
    cfg = OmegaConf.structured(User)
    a_before = cfg._get_node("age")
    cfg.age = 12
    assert id(cfg._get_node("age")) == id(a_before)
    with raises(ValidationError):
        cfg.age = []


def test_allow_objects() -> None:
    c = OmegaConf.create({"foo": AnyNode()})
    with raises(UnsupportedValueType):
        c.foo = IllegalType()
    c._set_flag("allow_objects", True)
    iv = IllegalType()
    c.foo = iv
    assert c.foo == iv


def test_dereference_missing() -> None:
    cfg = OmegaConf.create({"x": "???"})
    x_node = cfg._get_node("x")
    assert isinstance(x_node, Node)
    assert x_node._dereference_node() is x_node


@mark.parametrize(
    "make_func",
    [
        StringNode,
        IntegerNode,
        FloatNode,
        BooleanNode,
        BytesNode,
        PathNode,
        lambda val, is_optional: EnumNode(
            enum_type=Color, value=val, is_optional=is_optional
        ),
    ],
)
def test_validate_and_convert_none(make_func: Any) -> None:
    node = make_func("???", is_optional=False)
    ref_type_str = type_str(node._metadata.ref_type)
    with raises(
        ValidationError,
        match=re.escape(
            f"Incompatible value 'None' for field of type '{ref_type_str}'"
        ),
    ):
        node.validate_and_convert(None)


def test_dereference_interpolation_to_missing() -> None:
    cfg = OmegaConf.create({"x": "${y}", "y": "???"})
    x_node = cfg._get_node("x")
    assert isinstance(x_node, Node)
    assert x_node._maybe_dereference_node() is None
    with raises(InterpolationToMissingValueError):
        cfg.x


@mark.parametrize(
    "flags",
    [
        {},
        {"flag": True},
        {"flag": False},
        {"flag1": True, "flag2": False},
    ],
)
@mark.parametrize(
    "type_",
    [
        AnyNode,
        BooleanNode,
        BytesNode,
        functools.partial(EnumNode, enum_type=Color),
        FloatNode,
        IntegerNode,
        InterpolationResultNode,
        StringNode,
        PathNode,
    ],
)
def test_set_flags_in_init(type_: Any, flags: Dict[str, bool]) -> None:
    node = type_(value=None, flags=flags)
    for f, v in flags.items():
        assert node._get_flag(f) is v


@mark.parametrize(
    "flags",
    [
        None,
        {"flag": True},
        {"flag": False},
        {"readonly": True},
        {"readonly": False},
        {"flag1": True, "flag2": False, "readonly": False},
        {"flag1": False, "flag2": True, "readonly": True},
    ],
)
def test_interpolation_result_readonly(flags: Any) -> None:
    readonly = None if flags is None else flags.get("readonly")
    expected = [] if flags is None else list(flags.items())
    node = InterpolationResultNode("foo", flags=flags)

    # Check that flags are set to their desired value.
    for k, v in expected:
        assert node._get_node_flag(k) is v

    # If no value was provided for the "readonly" flag, it should be set.
    if readonly is None:
        assert node._get_node_flag("readonly")
